from buz.event.infrastructure.kombu.consume_strategy.consume_strategy import ConsumeStrategy
from buz.event.infrastructure.kombu.consume_strategy.queue_per_subscriber_consume_strategy import (
    QueuePerSubscriberConsumeStrategy,
)

__all__ = ["ConsumeStrategy", "QueuePerSubscriberConsumeStrategy"]



from __future__ import annotations

import json

from buz.kafka.infrastructure.deserializers.byte_deserializer import ByteDeserializer


class JSONByteDeserializer(ByteDeserializer[dict]):
    __STRING_ENCODING = "utf-8"

    def deserialize(self, data: bytes) -> dict:
        decoded_string = data.decode(self.__STRING_ENCODING)
        return json.loads(decoded_string)



from abc import abstractmethod
from typing import TypeVar

from buz.kafka.infrastructure.deserializers.byte_deserializer import ByteDeserializer
from buz.message import Message

T = TypeVar("T", bound=Message)


class BytesToMessageDeserializer(ByteDeserializer[T]):
    @abstractmethod
    def deserialize(self, data: bytes) -> T:
        pass



from abc import ABC
from typing import Generic, TypeVar

from buz.event.event import Event
from buz.event.async_subscriber import AsyncSubscriber
from buz.event.meta_base_subscriber import MetaBaseSubscriber

TEvent = TypeVar("TEvent", bound=Event)


class BaseAsyncSubscriber(Generic[TEvent], AsyncSubscriber[TEvent], MetaBaseSubscriber[TEvent], ABC):
    pass



from abc import abstractmethod
from typing import Callable

from buz.event import Event, Subscriber
from buz.event.infrastructure.models.execution_context import ExecutionContext
from buz.middleware import Middleware

ConsumeCallable = Callable[[Event, Subscriber, ExecutionContext], None]


class ConsumeMiddleware(Middleware):
    @abstractmethod
    def on_consume(
        self, event: Event, subscriber: Subscriber, consume: ConsumeCallable, execution_context: ExecutionContext
    ) -> None:
        pass



from typing import Optional
from buz.event.infrastructure.buz_kafka.consume_strategy.consume_strategy import KafkaConsumeStrategy
from buz.event.meta_subscriber import MetaSubscriber


class TopicAndSubscriptionGroupPerSubscriberKafkaConsumerStrategy(KafkaConsumeStrategy):
    def __init__(self, prefix: Optional[str]):
        self._prefix = f"{prefix}." or ""

    def get_topics(self, subscriber: MetaSubscriber) -> list[str]:
        event_class = subscriber.handles()
        return [f"{self._prefix}{event_class.fqn()}"]

    def get_subscription_group(self, subscriber: MetaSubscriber) -> str:
        return subscriber.fqn()



from __future__ import annotations

from ssl import SSLContext
from typing import Generic, Optional, TypeVar

from aiokafka import AIOKafkaProducer as NativeAIOKafkaProducer
from aiokafka.helpers import create_ssl_context

from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.models.kafka_supported_compression_type import KafkaSupportedCompressionType
from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols
from buz.kafka.domain.services.async_kafka_producer import AsyncKafkaProducer
from buz.kafka.infrastructure.serializers.byte_serializer import ByteSerializer
from buz.kafka.infrastructure.serializers.kafka_header_serializer import KafkaHeaderSerializer

T = TypeVar("T")


class AIOKafkaProducer(AsyncKafkaProducer, Generic[T]):
    __DEFAULT_REQUEST_TIMEOUT_MS = 5000
    __kafka_producer: Optional[NativeAIOKafkaProducer] = None

    def __init__(
        self,
        *,
        connection_config: KafkaConnectionConfig,
        byte_serializer: ByteSerializer[T],
        compression_type: Optional[KafkaSupportedCompressionType] = None,
        retry_backoff_ms: int = 100,
    ) -> None:
        self.__connection_config = connection_config
        self.__byte_serializer = byte_serializer
        self.__header_serializer = KafkaHeaderSerializer()
        self.__compression_type = compression_type
        self.__retry_backoff_ms = retry_backoff_ms

    async def _get_aiokafka_producer(self) -> NativeAIOKafkaProducer:
        if self.__kafka_producer:
            return self.__kafka_producer

        ssl_context: Optional[SSLContext] = None

        sasl_mechanism = (
            self.__connection_config.credentials.sasl_mechanism.value
            if self.__connection_config.credentials.sasl_mechanism
            else "PLAIN"
        )

        if self.__connection_config.credentials.security_protocol == KafkaSupportedSecurityProtocols.SASL_SSL:
            ssl_context = create_ssl_context()

        self.__kafka_producer = NativeAIOKafkaProducer(
            client_id=self.__connection_config.client_id,
            bootstrap_servers=",".join(self.__connection_config.bootstrap_servers),
            sasl_mechanism=sasl_mechanism,
            ssl_context=ssl_context,
            sasl_plain_username=self.__connection_config.credentials.user,
            sasl_plain_password=self.__connection_config.credentials.password,
            retry_backoff_ms=self.__retry_backoff_ms,
            request_timeout_ms=self.__DEFAULT_REQUEST_TIMEOUT_MS,
            compression_type=self.__compression_type.value if self.__compression_type else None,
        )

        await self.__kafka_producer.start()

        return self.__kafka_producer

    async def connect(self) -> None:
        await self._get_aiokafka_producer()

    async def disconnect(self) -> None:
        if self.__kafka_producer is None:
            return None
        await self.__kafka_producer.stop()
        self.__kafka_producer = None

    async def produce(
        self,
        *,
        topic: str,
        message: T,
        partition_key: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
    ) -> None:
        serialized_headers = self.__header_serializer.serialize(headers) if headers is not None else None
        kafka_producer = await self._get_aiokafka_producer()

        await kafka_producer.send_and_wait(
            topic=topic,
            value=self.__byte_serializer.serialize(message),
            headers=serialized_headers,
            key=partition_key.encode("utf-8") if partition_key else None,
        )

    async def close(self) -> None:
        if self.__kafka_producer is not None:
            await self.__kafka_producer.stop()
            self.__kafka_producer = None



from __future__ import annotations

from datetime import datetime
from typing import Optional, TypeVar, Type, Generic

import orjson
from dacite import from_dict

from buz.event import Event
from buz.kafka.infrastructure.cdc.cdc_message import CDCPayload
from buz.kafka.infrastructure.deserializers.bytes_to_message_deserializer import BytesToMessageDeserializer
from buz.kafka.infrastructure.deserializers.implementations.cdc.not_valid_cdc_message_exception import (
    NotValidCDCMessageException,
)

T = TypeVar("T", bound=Event)


class CDCRecordBytesToEventDeserializer(BytesToMessageDeserializer[Event], Generic[T]):
    __STRING_ENCODING = "utf-8"

    def __init__(self, event_class: Type[T]) -> None:
        self.__event_class = event_class

    def deserialize(self, data: bytes) -> T:
        decoded_string = data.decode(self.__STRING_ENCODING)
        try:
            cdc_payload = self.__get_outbox_record_as_dict(decoded_string)
            return self.__event_class.restore(
                id=cdc_payload.event_id,
                created_at=self.__get_created_at_in_event_format(cdc_payload.created_at),
                metadata=self.__deserialize_event_metadata(cdc_payload.event_metadata),
                **orjson.loads(cdc_payload.payload),
            )
        except Exception as exception:
            raise NotValidCDCMessageException(decoded_string, exception) from exception

    def __get_created_at_in_event_format(self, cdc_payload_created_at: str) -> str:
        created_at_datetime = datetime.strptime(cdc_payload_created_at, CDCPayload.DATE_TIME_FORMAT)
        return created_at_datetime.strftime(Event.DATE_TIME_FORMAT)

    def __get_outbox_record_as_dict(self, decoded_string: str) -> CDCPayload:
        decoded_record: dict = orjson.loads(decoded_string)

        payload = decoded_record.get("payload")

        if not isinstance(payload, dict):
            raise ValueError("The provided payload value is not valid")

        return from_dict(CDCPayload, payload)

    def __deserialize_event_metadata(self, event_metadata: Optional[str]) -> dict:
        if event_metadata is None:
            return {}
        return orjson.loads(event_metadata)



from typing import Type, Collection, Optional
from buz.event.event import Event
from buz.locator import MessageFqnNotFoundException

EventFqn = str


class FqnToEventMapper:
    def __init__(self, events: Collection[Type[Event]]):
        self.__events = events
        self.__events_map: Optional[dict[EventFqn, Type[Event]]] = None

    def get_message_klass_by_fqn(self, fqn: EventFqn) -> Type[Event]:
        self.__check_events_map_resolved()

        try:
            return self.__events_map[fqn]  # type: ignore[index]
        except KeyError:
            raise MessageFqnNotFoundException(fqn)

    def __check_events_map_resolved(self) -> None:
        if self.__events_map is None:
            self.__events_map = {event.fqn(): event for event in self.__events}



import asyncio
from logging import Logger
from typing import Optional, Callable, cast

from buz.event.infrastructure.kombu.models.kombu_delivery_context import KombuDeliveryContext
from buz.event.infrastructure.models.execution_context import ExecutionContext
from kombu import Connection, Queue, Consumer as MessageConsumer, Message
from kombu.mixins import ConsumerMixin
from kombu.transport.pyamqp import Channel

from buz.event import Event
from buz.event.async_subscriber import AsyncSubscriber
from buz.event.infrastructure.kombu.allowed_kombu_serializer import AllowedKombuSerializer
from buz.event.exceptions.term_signal_interruption_exception import TermSignalInterruptionException
from buz.event.meta_subscriber import MetaSubscriber
from buz.event.middleware import ConsumeMiddleware, ConsumeMiddlewareChainResolver
from buz.event.middleware.exceptions.event_already_in_progress_exception import EventAlreadyInProgressException
from buz.event.subscriber import Subscriber
from buz.locator import Locator
from buz.event.consumer import Consumer
from buz.event.exceptions.event_restore_exception import EventRestoreException
from buz.event.exceptions.subscribers_not_found_exception import SubscribersNotFoundException
from buz.event.strategies.retry.consume_retrier import ConsumeRetrier
from buz.event.strategies.retry.reject_callback import RejectCallback

QueueToSubscriberFqnMapping = dict[Queue, set[str]]


class KombuConsumer(ConsumerMixin, Consumer):
    def __init__(
        self,
        connection: Connection,
        queues_mapping: QueueToSubscriberFqnMapping,
        serializer: AllowedKombuSerializer,
        prefetch_count: int,
        locator: Locator[Event, MetaSubscriber],
        logger: Logger,
        consume_retrier: Optional[ConsumeRetrier] = None,
        reject_callback: Optional[RejectCallback] = None,
        consume_middlewares: Optional[list[ConsumeMiddleware]] = None,
    ):
        self.connection = connection
        self.__queues_mapping = queues_mapping
        self.__serializer = serializer
        self.__prefetch_count = prefetch_count
        self.__locator = locator
        self.__logger = logger
        self.__consume_retrier = consume_retrier
        self.__reject_callback = reject_callback
        self.__consume_middleware_chain_resolver = ConsumeMiddlewareChainResolver(consume_middlewares or [])

    def stop(self) -> None:
        self.should_stop = True

    def get_consumers(self, consumer_factory: Callable, channel: Channel) -> list[MessageConsumer]:
        return [
            consumer_factory(
                queues=[queue],
                callbacks=self.__get_consumer_callbacks(allowed_subscriber_fqns),
                prefetch_count=self.__prefetch_count,
                accept=[self.__serializer],
            )
            for queue, allowed_subscriber_fqns in self.__queues_mapping.items()
        ]

    def __get_consumer_callbacks(self, allowed_subscriber_fqns: set[str]) -> list[Callable[[dict, Message], None]]:
        return [lambda body, message: self.__on_message_received(body, message, allowed_subscriber_fqns)]

    def __on_message_received(self, body: dict, message: Message, allowed_subscriber_fqns: set[str]) -> None:
        try:
            event = self.__restore_event(body, message)
            subscribers = self.__subscribers(event, allowed_subscriber_fqns)
        except (EventRestoreException, SubscribersNotFoundException) as exc:
            self.__logger.exception(f"Message cannot be processed: {exc}")
            message.ack()
            return
        except Exception as exc:
            self.__logger.exception(f"Unknown error while processing message: {exc}")
            message.ack()
            return

        self.__consume_event(message, event, subscribers)

    def __restore_event(self, body: dict, message: Message) -> Event:
        try:
            event_fqn = message.headers["fqn"]
            event_klass = self.__locator.get_message_klass_by_fqn(event_fqn)
            return cast(Event, event_klass.restore(**body))
        except Exception as exc:
            raise EventRestoreException(body, str(message)) from exc

    def __subscribers(self, event: Event, allowed_subscriber_fqns: set[str]) -> list[MetaSubscriber]:
        event_subscribers = None
        try:
            event_subscribers = self.__locator.get(event)
            allowed_event_subscribers = [
                subscriber for subscriber in event_subscribers if subscriber.fqn() in allowed_subscriber_fqns
            ]
            if len(allowed_event_subscribers) == 0:
                raise SubscribersNotFoundException(event, allowed_subscriber_fqns, event_subscribers)

            return allowed_event_subscribers
        except Exception as exc:
            raise SubscribersNotFoundException(event, allowed_subscriber_fqns, event_subscribers) from exc

    def __consume_event(self, message: Message, event: Event, meta_subscribers: list[MetaSubscriber]) -> None:
        try:
            if self.should_stop is True:
                raise TermSignalInterruptionException()

            for meta_subscriber in meta_subscribers:
                # The problem here is that the chain resolver works with syncsubscribers, an asyncsubscriber would require of a async function,
                # but we are using run-until-complete to run the async function, so we are not really using the async function, we are just running it as a sync function, so we can cast the asyncsubscriber to a subscriber
                subscriber = cast(Subscriber, meta_subscriber)
                execution_context = ExecutionContext(delivery_context=KombuDeliveryContext())
                self.__consume_middleware_chain_resolver.resolve(
                    event=event,
                    subscriber=subscriber,
                    execution_context=execution_context,
                    consume=self.__perform_consume,
                )
            message.ack()
        except Exception as exc:
            self.__on_consume_exception(message, event, meta_subscribers, exc)

    def __perform_consume(self, event: Event, subscriber: MetaSubscriber, execution_context: ExecutionContext) -> None:
        if isinstance(subscriber, AsyncSubscriber):
            self.__get_or_create_event_loop().run_until_complete(subscriber.consume(event))
            return

        if isinstance(subscriber, Subscriber):
            return subscriber.consume(event)

        raise TypeError(f"Subscriber {subscriber} is not an instance of Subscriber or AsyncSubscriber")

    def __get_or_create_event_loop(self) -> asyncio.AbstractEventLoop:
        try:
            return asyncio.get_event_loop()
        except RuntimeError:
            self.__logger.warning("The current thread does not have an event loop attached. Creating new event loop.")
            event_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(event_loop)
            return event_loop

    def __on_consume_exception(
        self, message: Message, event: Event, subscribers: list[MetaSubscriber], exception: Exception
    ) -> None:
        if isinstance(exception, EventAlreadyInProgressException) or isinstance(
            exception, TermSignalInterruptionException
        ):
            message.requeue()
            return

        self.__logger.warning(
            f"Event {event.id} could not be consumed by subscribers {[subscriber.fqn() for subscriber in subscribers]}:"
            f"{exception}."
        )

        if self.__consume_retrier is None:
            self.__reject_message(message, event, subscribers, exception)
            return

        should_retry = self.__consume_retrier.should_retry(event, subscribers)
        if should_retry is True:
            self.__consume_retrier.register_retry(event, subscribers)
            message.requeue()
            return

        self.__logger.exception(exception)
        self.__reject_message(message, event, subscribers, exception)

    def __reject_message(
        self, message: Message, event: Event, subscribers: list[MetaSubscriber], exception: Exception
    ) -> None:
        message.reject()
        if self.__reject_callback is not None:
            self.__reject_callback.on_reject(event=event, subscribers=subscribers, exception=exception)



from buz.query import QueryResponse, Query
from buz.query.synchronous import QueryBus
from buz.query.asynchronous import QueryBus as AsynchronousQueryBus
from buz.wrapper import AsyncToSync


class SyncedAsyncQueryBus(QueryBus):
    def __init__(self, async_bus: AsynchronousQueryBus):
        self.__wrapped_handler = AsyncToSync[Query, QueryResponse](async_bus.handle)

    def handle(self, query: Query) -> QueryResponse:
        return self.__wrapped_handler(query)



from typing import Optional

import orjson
from pympler.asizeof import asizeof

from buz.event.transactional_outbox import OutboxRecord
from buz.event.transactional_outbox.outbox_record_validation.abstract_outbox_record_validator import (
    AbstractOutboxRecordValidator,
)
from buz.event.transactional_outbox.outbox_record_validation.outbox_record_size_not_allowed_exception import (
    OutboxRecordSizeNotAllowedException,
)


class SizeOutboxRecordValidator(AbstractOutboxRecordValidator):
    def __init__(self, size_limit_in_bytes: int = 1000000):
        self.__size_limit_in_bytes = size_limit_in_bytes
        super().__init__()

    def validate(self, record: OutboxRecord) -> None:
        try:
            size = asizeof(self.__record_to_json(record))
        except Exception:
            size = self.__fallback_size_validator(record)

        if size >= self.__size_limit_in_bytes:
            raise OutboxRecordSizeNotAllowedException(
                record=record, size_limit_in_bytes=self.__size_limit_in_bytes, record_size=size
            )

        return super().validate(record)

    def __record_to_json(self, record: OutboxRecord) -> bytes:
        [created_at, delivered_at, delivery_paused_at] = self.__get_dates_parsed_for_json(record)
        return orjson.dumps(
            {
                "created_at": created_at,
                "delivered_at": delivered_at,
                "delivery_paused_at": delivery_paused_at,
                "delivery_errors": record.delivery_errors,
                "event_id": record.event_id,
                "event_fqn": record.event_fqn,
                "event_payload": record.event_payload,
            }
        )

    def __get_dates_parsed_for_json(self, record: OutboxRecord) -> tuple[str, Optional[str], Optional[str]]:
        parsed_created_at_json = record.created_at.strftime(record.DATE_TIME_FORMAT)
        parsed_delivered_at_json = (
            record.delivered_at.strftime(record.DATE_TIME_FORMAT) if record.delivered_at is not None else None
        )
        parsed_delivery_paused_at_json = (
            record.delivery_paused_at.strftime(record.DATE_TIME_FORMAT)
            if record.delivery_paused_at is not None
            else None
        )
        return parsed_created_at_json, parsed_delivered_at_json, parsed_delivery_paused_at_json

    def __fallback_size_validator(self, record: OutboxRecord) -> int:
        return asizeof(record)



from buz.kafka.domain.exceptions.not_all_partition_assigned_exception import NotAllPartitionAssignedException
from buz.kafka.domain.exceptions.not_valid_kafka_message_exception import NotValidKafkaMessageException
from buz.kafka.domain.exceptions.topic_already_created_exception import KafkaTopicsAlreadyCreatedException
from buz.kafka.domain.exceptions.topic_not_found_exception import TopicNotFoundException
from buz.kafka.domain.models.auto_create_topic_configuration import AutoCreateTopicConfiguration
from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition
from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.models.kafka_connection_credentials import KafkaConnectionCredentials
from buz.kafka.domain.models.kafka_connection_plain_text_credentials import KafkaConnectionPlainTextCredentials
from buz.kafka.domain.models.kafka_connection_sasl_credentials import KafkaConnectionSaslCredentials
from buz.kafka.domain.models.kafka_consumer_record import KafkaConsumerRecord
from buz.kafka.domain.models.kafka_supported_sasl_mechanisms import KafkaSupportedSaslMechanisms
from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols
from buz.kafka.domain.models.create_kafka_topic import CreateKafkaTopic
from buz.kafka.domain.services.kafka_admin_client import KafkaAdminClient
from buz.kafka.domain.services.kafka_admin_test_client import KafkaAdminTestClient
from buz.kafka.domain.services.kafka_producer import KafkaProducer
from buz.kafka.infrastructure.aiokafka.aiokafka_producer import AIOKafkaProducer
from buz.kafka.infrastructure.kafka_python.kafka_python_admin_client import KafkaPythonAdminClient
from buz.kafka.infrastructure.kafka_python.kafka_python_admin_test_client import KafkaPythonAdminTestClient
from buz.kafka.infrastructure.kafka_python.kafka_python_producer import KafkaPythonProducer
from buz.kafka.infrastructure.serializers.byte_serializer import ByteSerializer
from buz.kafka.infrastructure.serializers.implementations.json_byte_serializer import JSONByteSerializer
from buz.kafka.domain.models.kafka_supported_compression_type import KafkaSupportedCompressionType
from buz.event.infrastructure.buz_kafka.exceptions.kafka_event_bus_config_not_valid_exception import (
    KafkaEventBusConfigNotValidException,
)
from buz.event.infrastructure.buz_kafka.async_buz_kafka_event_bus import AsyncBuzKafkaEventBus
from buz.event.infrastructure.buz_kafka.buz_kafka_event_bus import BuzKafkaEventBus


__all__ = [
    "KafkaProducer",
    "KafkaPythonProducer",
    "KafkaAdminClient",
    "KafkaAdminTestClient",
    "KafkaPythonAdminClient",
    "KafkaPythonAdminTestClient",
    "KafkaTopicsAlreadyCreatedException",
    "KafkaConsumerRecord",
    "CreateKafkaTopic",
    "KafkaSupportedSecurityProtocols",
    "KafkaConnectionConfig",
    "ByteSerializer",
    "JSONByteSerializer",
    "ConsumerInitialOffsetPosition",
    "KafkaSupportedCompressionType",
    "KafkaEventBusConfigNotValidException",
    "AsyncBuzKafkaEventBus",
    "BuzKafkaEventBus",
    "AutoCreateTopicConfiguration",
    "NotAllPartitionAssignedException",
    "NotValidKafkaMessageException",
    "TopicNotFoundException",
    "KafkaConnectionCredentials",
    "KafkaConnectionPlainTextCredentials",
    "KafkaConnectionSaslCredentials",
    "KafkaSupportedSaslMechanisms",
    "AIOKafkaProducer",
]



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

T = TypeVar("T")


class ByteDeserializer(ABC, Generic[T]):
    @abstractmethod
    def deserialize(self, data: bytes) -> T:
        pass



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

T = TypeVar("T")


class ByteSerializer(ABC, Generic[T]):
    @abstractmethod
    def serialize(self, data: T) -> bytes:
        pass



class MaxConsumerRetryException(Exception):
    def __init__(
        self,
        *,
        event_id: str,
        subscriber_fqn: str,
    ) -> None:
        super().__init__(f"Max retries reached, discarding event {event_id} for subscriber {subscriber_fqn}")



class ConsumerRetryException(Exception):
    def __init__(
        self,
        *,
        event_id: str,
        subscriber_fqn: str,
        number_of_executions: int,
    ) -> None:
        super().__init__(
            f"An exception happened during the consumption of the event '{event_id}' by the subscriber '{subscriber_fqn}' "
            + f"during execution number '{number_of_executions}'. Retrying the consumption..."
        )



import asyncio
from logging import Logger
from signal import SIGTERM, SIGINT

from buz.event.strategies.execution_strategy.async_execution_strategy import AsyncExecutionStrategy


class AsyncWorker:
    def __init__(
        self,
        logger: Logger,
        execution_strategy: AsyncExecutionStrategy,
    ) -> None:
        self.__execution_strategy = execution_strategy
        self.__logger = logger

    async def start(self) -> None:
        loop = asyncio.get_running_loop()
        loop.add_signal_handler(SIGINT, self.stop)
        loop.add_signal_handler(SIGTERM, self.stop)
        self.__logger.info("Starting buz worker...")
        await self.__execution_strategy.start()
        self.__logger.info("Buz worker stopped gracefully")

    def stop(self) -> None:
        self.__logger.info("Stopping buz worker...")
        self.__execution_strategy.request_stop()



from buz.event import Event


class EventNotPublishedException(Exception):
    def __init__(self, event: Event) -> None:
        self.event = event
        super().__init__(f"Event could not be published: {event}")



from buz.event import Event
from buz.event.middleware.async_publish_middleware import AsyncPublishCallable, AsyncPublishMiddleware
from buz.middleware import MiddlewareChainBuilder


class AsyncPublishMiddlewareChainResolver:
    def __init__(self, middlewares: list[AsyncPublishMiddleware]):
        self.__middlewares = middlewares
        self.__middleware_chain_builder: MiddlewareChainBuilder[
            AsyncPublishCallable, AsyncPublishMiddleware
        ] = MiddlewareChainBuilder(middlewares)

    async def resolve(self, event: Event, publish: AsyncPublishCallable) -> None:
        chain_callable: AsyncPublishCallable = self.__middleware_chain_builder.get_chain_callable(
            publish, self.__get_middleware_callable
        )
        await chain_callable(event)

    def __get_middleware_callable(
        self, middleware: AsyncPublishMiddleware, publish_callable: AsyncPublishCallable
    ) -> AsyncPublishCallable:
        return lambda event: middleware.on_publish(event, publish_callable)



from buz.event.async_consumer import AsyncConsumer
from buz.event.strategies.execution_strategy.async_execution_strategy import AsyncExecutionStrategy


class AsyncSelfProcessExecutionStrategy(AsyncExecutionStrategy):
    def __init__(self, consumer: AsyncConsumer):
        self.__consumer = consumer

    async def start(self) -> None:
        await self.__consumer.run()

    def request_stop(self) -> None:
        self.__consumer.request_stop()



from abc import abstractmethod

from buz.event import Event, Subscriber
from buz.event.infrastructure.models.execution_context import ExecutionContext
from buz.event.middleware import ConsumeMiddleware, ConsumeCallable


class BaseConsumeMiddleware(ConsumeMiddleware):
    def on_consume(
        self, event: Event, subscriber: Subscriber, consume: ConsumeCallable, execution_context: ExecutionContext
    ) -> None:
        self._before_on_consume(event, subscriber, execution_context)
        consume(event, subscriber, execution_context)
        self._after_on_consume(event, subscriber, execution_context)

    @abstractmethod
    def _before_on_consume(self, event: Event, subscriber: Subscriber, execution_context: ExecutionContext) -> None:
        pass

    @abstractmethod
    def _after_on_consume(self, event: Event, subscriber: Subscriber, execution_context: ExecutionContext) -> None:
        pass



class NotValidPartitionNumberException(Exception):
    def __init__(
        self,
        *,
        partition_number: int,
        min_partition_number: int,
    ) -> None:
        super().__init__(
            f'"{partition_number}" is not a valid partition number, the minimum partition number is "{min_partition_number}"'
        )



from abc import ABC, abstractmethod
from typing import Iterable

from buz.event.transactional_outbox import OutboxRecord


class OutboxRecordStreamFinder(ABC):
    @abstractmethod
    def find(self) -> Iterable[OutboxRecord]:
        pass



from buz.query.synchronous.base_query_handler import BaseQueryHandler
from buz.query.synchronous.query_bus import QueryBus
from buz.query.synchronous.query_handler import QueryHandler

__all__ = ["QueryHandler", "BaseQueryHandler", "QueryBus"]



from typing import Generic, Type, TypeVar, cast, get_type_hints, Any

from buz.query import Query
from buz.query.asynchronous.query_handler import QueryHandler
from buz.query.query_response import QueryResponse

TQuery = TypeVar("TQuery", bound=Query)
TQueryResponse = TypeVar("TQueryResponse", bound=QueryResponse)


class BaseQueryHandler(Generic[TQuery, TQueryResponse], QueryHandler[TQuery, TQueryResponse]):
    @classmethod
    def fqn(cls) -> str:
        return f"query_handler.{cls.__module__}.{cls.__name__}"

    @classmethod
    def handles(cls) -> Type[TQuery]:
        handle_types = get_type_hints(cls.handle)

        t_query = handle_types.get("query")
        if t_query is None:
            raise TypeError(
                f"The method 'handle' in '{cls.fqn()}' doesn't have a parameter named 'query'. Found parameters: {cls.__get_method_parameter_names(handle_types)}"
            )

        # TypeVar mask the actual bound type
        if hasattr(t_query, "__bound__"):
            t_query = t_query.__bound__

        if not issubclass(t_query, Query):
            raise TypeError(f"The parameter 'query' in '{cls.fqn()}.handle' is not a 'buz.query.Query' subclass")

        return cast(Type[TQuery], t_query)

    @classmethod
    def __get_method_parameter_names(cls, handle_types: dict[str, Any]) -> list[str]:
        handle_types_copy: dict = handle_types.copy()
        handle_types_copy.pop("return", None)
        return list(handle_types_copy.keys())



from abc import ABC

from buz.kafka.domain.models.kafka_poll_record import KafkaPollRecord


class KafkaEventSubscriberExecutor(ABC):
    async def consume(
        self,
        *,
        kafka_poll_record: KafkaPollRecord,
    ) -> None:
        pass



from typing import Optional, Sequence

from buz.event import Event
from buz.event.meta_subscriber import MetaSubscriber


class SubscribersNotFoundException(Exception):
    def __init__(
        self,
        event: Event,
        allowed_subscriber_fqns: set[str],
        event_subscribers: Optional[Sequence[MetaSubscriber]] = None,
    ) -> None:
        self.event = event
        self.allowed_subscriber_fqns = allowed_subscriber_fqns
        self.event_subscribers = event_subscribers

        message = f"Subscribers not found for event {event}. Allowed subscriber FQNs: {allowed_subscriber_fqns}."
        if event_subscribers is not None:
            message += f" All event subscribers found: {event_subscribers}"

        super().__init__(message)



from __future__ import annotations

from dataclasses import dataclass

from buz.kafka.domain.models.kafka_connection_credentials import KafkaConnectionCredentials


@dataclass(frozen=True)
class KafkaConnectionConfig:
    bootstrap_servers: list[str]
    client_id: str
    credentials: KafkaConnectionCredentials



from buz.kafka.domain.exceptions.not_valid_kafka_message_exception import NotValidKafkaMessageException


class NotValidCDCMessageException(NotValidKafkaMessageException):
    def __init__(self, message: str, exception: Exception) -> None:
        self.exception = exception
        self.message = message
        super().__init__(f'The CDC message "{self.message}" was not decoded correctly: "{str(exception)}"')



from typing import Sequence
from buz.event import Event
from buz.event.async_subscriber import AsyncSubscriber

from buz.event.infrastructure.models.execution_context import ExecutionContext
from buz.event.middleware.async_consume_middleware import AsyncConsumeCallable, AsyncConsumeMiddleware
from buz.middleware import MiddlewareChainBuilder


class AsyncConsumeMiddlewareChainResolver:
    def __init__(
        self,
        middlewares: Sequence[AsyncConsumeMiddleware],
    ):
        self.__middlewares = middlewares
        self.__middleware_chain_builder: MiddlewareChainBuilder[
            AsyncConsumeCallable, AsyncConsumeMiddleware
        ] = MiddlewareChainBuilder(self.__middlewares)

    async def resolve(
        self,
        event: Event,
        subscriber: AsyncSubscriber,
        consume: AsyncConsumeCallable,
        execution_context: ExecutionContext,
    ) -> None:
        chain_callable: AsyncConsumeCallable = self.__middleware_chain_builder.get_chain_callable(
            consume, self.__get_middleware_callable
        )

        await chain_callable(event, subscriber, execution_context)

    def __get_middleware_callable(
        self,
        middleware: AsyncConsumeMiddleware,
        consume_callable: AsyncConsumeCallable,
    ) -> AsyncConsumeCallable:
        return lambda event, subscriber, execution_context: middleware.on_consume(
            event, subscriber, consume_callable, execution_context
        )



from abc import abstractmethod
from typing import Awaitable, Callable

from buz.event import Event
from buz.middleware import Middleware

AsyncPublishCallable = Callable[[Event], Awaitable[None]]


class AsyncPublishMiddleware(Middleware):
    @abstractmethod
    async def on_publish(self, event: Event, publish: AsyncPublishCallable) -> None:
        pass



from buz.event.event import Event
from buz.event.subscriber import Subscriber
from buz.event.async_subscriber import AsyncSubscriber
from buz.event.base_subscriber import BaseSubscriber
from buz.event.base_async_subscriber import BaseAsyncSubscriber
from buz.event.event_bus import EventBus
from buz.event.meta_subscriber import MetaSubscriber
from buz.event.meta_base_subscriber import MetaBaseSubscriber

__all__ = [
    "Event",
    "Subscriber",
    "AsyncSubscriber",
    "BaseSubscriber",
    "BaseAsyncSubscriber",
    "EventBus",
    "MetaSubscriber",
    "MetaBaseSubscriber",
]



from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar

K = TypeVar("K")
R = TypeVar("R")


class MultiqueueRepository(ABC, Generic[K, R]):
    @abstractmethod
    def create_key(self, key: K) -> None:
        pass

    @abstractmethod
    def remove_key(self, key: K) -> None:
        pass

    @abstractmethod
    def push(self, key: K, record: R) -> None:
        pass

    @abstractmethod
    def pop(self) -> Optional[R]:
        pass

    @abstractmethod
    def get_total_size(self) -> int:
        pass

    @abstractmethod
    def is_totally_empty(self) -> bool:
        pass



from typing import Generic, Type, TypeVar, cast, get_type_hints, Any

from buz.command import Command
from buz.command.asynchronous.command_handler import CommandHandler

TCommand = TypeVar("TCommand", bound=Command)


class BaseCommandHandler(Generic[TCommand], CommandHandler[TCommand]):
    @classmethod
    def fqn(cls) -> str:
        return f"command_handler.{cls.__module__}.{cls.__name__}"

    @classmethod
    def handles(cls) -> Type[TCommand]:
        handle_types = get_type_hints(cls.handle)

        t_command = handle_types.get("command")
        if t_command is None:
            raise TypeError(
                f"The method 'handle' in '{cls.fqn()}' doesn't have a parameter named 'command'. Found parameters: {cls.__get_method_parameter_names(handle_types)}"
            )

        # TypeVar mask the actual bound type
        if hasattr(t_command, "__bound__"):
            t_command = t_command.__bound__

        if not issubclass(t_command, Command):
            raise TypeError(f"The parameter 'command' in '{cls.fqn()}.handle' is not a 'buz.command.Command' subclass")

        return cast(Type[TCommand], t_command)

    @classmethod
    def __get_method_parameter_names(cls, handle_types: dict[str, Any]) -> list[str]:
        handle_types_copy: dict = handle_types.copy()
        handle_types_copy.pop("return", None)
        return list(handle_types_copy.keys())



from abc import ABC
from typing import Generic, Type, TypeVar, cast, get_type_hints

from buz.event.event import Event
from buz.event.meta_subscriber import MetaSubscriber

TEvent = TypeVar("TEvent", bound=Event)


class MetaBaseSubscriber(Generic[TEvent], MetaSubscriber[TEvent], ABC):
    @classmethod
    def fqn(cls) -> str:
        return f"subscriber.{cls.__module__}.{cls.__name__}"

    @classmethod
    def handles(cls) -> Type[TEvent]:
        consume_types = get_type_hints(cls.consume)

        t_event = consume_types.get("event")
        if t_event is None:
            raise TypeError("event parameter not found in consume method")

        if not issubclass(t_event, Event):
            raise TypeError("event parameter is not an buz.event.Event subclass")

        return cast(Type[TEvent], t_event)



from collections import defaultdict
import json
import traceback
from abc import abstractmethod
from asyncio import Lock, Task, create_task, gather, Semaphore, Event as AsyncIOEvent, sleep
from datetime import timedelta, datetime
from itertools import cycle
from logging import Logger
from typing import AsyncIterator, Optional, Sequence, Type, TypeVar
from aiohttp import web

from aiokafka import TopicPartition
from aiokafka.coordinator.assignors.abstract import AbstractPartitionAssignor
from aiokafka.coordinator.assignors.roundrobin import RoundRobinPartitionAssignor

from buz.event import Event
from buz.event.async_consumer import AsyncConsumer
from buz.event.exceptions.worker_execution_exception import WorkerExecutionException
from buz.event.infrastructure.buz_kafka.consume_strategy.consume_strategy import KafkaConsumeStrategy
from buz.event.infrastructure.buz_kafka.kafka_event_subscriber_executor import KafkaEventSubscriberExecutor
from buz.event.infrastructure.models.consuming_task import ConsumingTask
from buz.event.meta_subscriber import MetaSubscriber
from buz.kafka.domain.models.auto_create_topic_configuration import AutoCreateTopicConfiguration
from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition
from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.models.kafka_poll_record import KafkaPollRecord
from buz.kafka.domain.services.kafka_admin_client import KafkaAdminClient
from buz.kafka.infrastructure.aiokafka.aiokafka_consumer import AIOKafkaConsumer
from buz.queue.in_memory.in_memory_multiqueue_repository import InMemoryMultiqueueRepository
from buz.queue.multiqueue_repository import MultiqueueRepository

T = TypeVar("T", bound=Event)


class BaseBuzAIOKafkaAsyncConsumer(AsyncConsumer):
    __FALLBACK_PARTITION_ASSIGNORS = (RoundRobinPartitionAssignor,)
    __DEFAULT_MAX_POLL_INTERVAL = 30 * 60 * 1000
    __DEFAULT_SESSION_TIMEOUT_MS = 1000 * 120
    __DEFAULT_HEARTBEAT_INTERVAL = int(__DEFAULT_SESSION_TIMEOUT_MS / 5)
    __SECONDS_BETWEEN_EXECUTIONS_IF_THERE_ARE_NO_TASKS_IN_THE_QUEUE = 1
    __SECONDS_BETWEEN_POLLS_IF_THERE_ARE_TASKS_IN_THE_QUEUE = 1
    __SECONDS_BETWEEN_POLLS_IF_THERE_ARE_NO_NEW_TASKS = 1
    __SECONDS_TO_WAIT_BETWEEN_REBALANCING_IN_PROGRESS = 0.1
    __MAX_NUMBER_OF_CONCURRENT_POLLING_TASKS = 20

    def __init__(
        self,
        *,
        connection_config: KafkaConnectionConfig,
        kafka_admin_client: Optional[KafkaAdminClient],
        consume_strategy: KafkaConsumeStrategy,
        max_queue_size: int,
        max_records_retrieved_per_poll: int,
        kafka_partition_assignors: tuple[Type[AbstractPartitionAssignor], ...] = (),
        subscribers: Sequence[MetaSubscriber],
        logger: Logger,
        health_check_port: Optional[int],
        consumer_initial_offset_position: ConsumerInitialOffsetPosition,
        auto_create_topic_configuration: Optional[AutoCreateTopicConfiguration] = None,
        seconds_between_executions_if_there_are_no_tasks_in_the_queue: Optional[int] = None,
        seconds_between_polls_if_there_are_tasks_in_the_queue: Optional[int] = None,
        seconds_between_polls_if_there_are_no_new_tasks: Optional[int] = None,
        max_number_of_concurrent_polling_tasks: Optional[int] = None,
        session_timeout_ms: Optional[int] = None,
        max_poll_interval_ms: Optional[int] = None,
        heartbeat_interval_ms: Optional[int] = None,
        wait_for_connection_to_cluster_ms: Optional[int] = None,
        worker_instance_id: Optional[str] = None,
    ):
        self.__connection_config = connection_config
        self.__consume_strategy = consume_strategy
        self.__kafka_partition_assignors = kafka_partition_assignors
        self.__subscribers = subscribers
        self._logger = logger
        self.__health_check_port = health_check_port
        self.__consumer_initial_offset_position = consumer_initial_offset_position
        self.__executor_per_consumer_mapper: dict[AIOKafkaConsumer, KafkaEventSubscriberExecutor] = {}
        self.__queue_per_consumer_mapper: dict[
            AIOKafkaConsumer, MultiqueueRepository[TopicPartition, KafkaPollRecord]
        ] = {}
        self.__max_records_retrieved_per_poll: int = max_records_retrieved_per_poll
        self.__session_timeout_ms: int = session_timeout_ms or self.__DEFAULT_SESSION_TIMEOUT_MS
        self.__max_poll_interval_ms: int = max_poll_interval_ms or self.__DEFAULT_MAX_POLL_INTERVAL
        self.__heartbeat_interval_ms: int = heartbeat_interval_ms or self.__DEFAULT_HEARTBEAT_INTERVAL
        self.__max_queue_size: int = max_queue_size
        self.__should_stop: AsyncIOEvent = AsyncIOEvent()
        self.__start_kafka_consumers_elapsed_time: Optional[timedelta] = None
        self.__initial_coroutines_created_elapsed_time: Optional[timedelta] = None
        self.__events_processed: int = 0
        self.__events_processed_elapsed_time: timedelta = timedelta()
        self.__kafka_admin_client: Optional[KafkaAdminClient] = kafka_admin_client
        self.__auto_create_topic_configuration: Optional[AutoCreateTopicConfiguration] = auto_create_topic_configuration
        self.__seconds_between_executions_if_there_are_no_tasks_in_the_queue: int = (
            seconds_between_executions_if_there_are_no_tasks_in_the_queue
            or self.__SECONDS_BETWEEN_EXECUTIONS_IF_THERE_ARE_NO_TASKS_IN_THE_QUEUE
        )
        self.__seconds_between_polls_if_there_are_tasks_in_the_queue: int = (
            seconds_between_polls_if_there_are_tasks_in_the_queue
            or self.__SECONDS_BETWEEN_POLLS_IF_THERE_ARE_TASKS_IN_THE_QUEUE
        )
        self.__seconds_between_polls_if_there_are_no_new_tasks: int = (
            seconds_between_polls_if_there_are_no_new_tasks or self.__SECONDS_BETWEEN_POLLS_IF_THERE_ARE_NO_NEW_TASKS
        )
        self.__max_number_of_concurrent_polling_tasks: int = (
            max_number_of_concurrent_polling_tasks or self.__MAX_NUMBER_OF_CONCURRENT_POLLING_TASKS
        )
        self.__wait_for_connection_to_cluster_ms: Optional[int] = wait_for_connection_to_cluster_ms
        self.__worker_instance_id: Optional[str] = worker_instance_id
        self.__polling_tasks_semaphore = Semaphore(self.__max_number_of_concurrent_polling_tasks)
        self.__consumer_and_partition_mutex: dict[str, Lock] = defaultdict(Lock)
        self.__is_worked_initialized = False
        self.__number_of_rebalancing_processes_in_progress: int = 0

    async def configure_http_check_server(self, health_check_port: int) -> web.TCPSite:
        self._logger.info(f"Starting health check server on port {health_check_port}")
        app = web.Application()
        app.router.add_get("/health", self.__health_check)
        app.router.add_get("/ready", self.__is_ready)

        runner = web.AppRunner(app)
        await runner.setup()
        site = web.TCPSite(runner, "0.0.0.0", health_check_port)
        await site.start()
        return site

    async def run(self) -> None:
        self.__print_worker_configuration()
        start_time = datetime.now()
        health_check_server: Optional[web.TCPSite] = None

        if self.__health_check_port is not None:
            health_check_server = await self.configure_http_check_server(self.__health_check_port)

        await self.__generate_kafka_consumers()

        self.__initial_coroutines_created_elapsed_time = datetime.now() - start_time

        if len(self.__executor_per_consumer_mapper) == 0:
            self._logger.error("There are no valid subscribers to execute, finalizing consumer")
            return

        self.__is_worked_initialized = True

        start_consumption_time = datetime.now()
        worker_errors = await self.__run_worker()
        self.__events_processed_elapsed_time = datetime.now() - start_consumption_time

        await self.__handle_graceful_stop(worker_errors)

        if health_check_server is not None:
            await health_check_server.stop()

    def __print_worker_configuration(self) -> None:
        self._logger.info(
            f"Consumer configuration:\n"
            f"  - Worker instance id: {self.__worker_instance_id}\n"
            f"  - Consume strategy: {self.__consume_strategy}\n"
            f"  - Max queue size: {self.__max_queue_size}\n"
            f"  - Max records retrieved per poll: {self.__max_records_retrieved_per_poll}\n"
            f"  - Kafka partition assignors: {self.__kafka_partition_assignors}\n"
            f"  - Consumer initial offset position: {self.__consumer_initial_offset_position}\n"
            f"  - Session timeout ms: {self.__session_timeout_ms}\n"
            f"  - Max poll interval ms: {self.__max_poll_interval_ms}\n"
            f"  - Heartbeat interval ms: {self.__heartbeat_interval_ms}\n"
            f"  - Seconds between executions if there are no tasks in the queue: {self.__seconds_between_executions_if_there_are_no_tasks_in_the_queue}\n"
            f"  - Seconds between polls if there are tasks in the queue: {self.__seconds_between_polls_if_there_are_tasks_in_the_queue}\n"
            f"  - Seconds between polls if there are no new tasks: {self.__seconds_between_polls_if_there_are_no_new_tasks}\n"
            f"  - Max number of concurrent polling tasks: {self.__max_number_of_concurrent_polling_tasks}\n"
            f"  - Wait for connection to cluster ms: {self.__wait_for_connection_to_cluster_ms}\n"
            f"  - Health check port: {self.__health_check_port}\n"
            f"  - Number of subscribers: {len(self.__subscribers)}",
        )

    async def __handle_graceful_stop(self, worker_errors: tuple[Optional[Exception], Optional[Exception]]) -> None:
        self._logger.info("Stopping kafka consumers...")
        await self.__manage_kafka_consumers_stopping()
        self._logger.info("All kafka consumers stopped")

        self.__print_statistics()

        if self.__exceptions_are_thrown(worker_errors):
            consume_events_exception, polling_task_exception = worker_errors
            raise WorkerExecutionException(
                "The worker was closed by an unexpected exception"
            ) from consume_events_exception or polling_task_exception

    async def __run_worker(self) -> tuple[Optional[Exception], Optional[Exception]]:
        consume_events_task = create_task(self.__consume_events_task())
        polling_task = create_task(self.__polling_task())

        try:
            await gather(consume_events_task, polling_task)
            return (None, None)
        except Exception:
            self.__should_stop.set()
            consume_events_exception = await self.__await_exception(consume_events_task)
            polling_task_exception = await self.__await_exception(polling_task)
            return (consume_events_exception, polling_task_exception)

    async def __await_exception(self, task: Task) -> Optional[Exception]:
        try:
            await task
            return None
        except Exception as exception:
            return exception

    def __exceptions_are_thrown(self, worker_errors: tuple[Optional[Exception], Optional[Exception]]) -> bool:
        return any([error is not None for error in worker_errors])

    async def __generate_kafka_consumers(self):
        start_time = datetime.now()
        tasks = [self.__generate_kafka_consumer_for_subscriber(subscriber) for subscriber in self.__subscribers]
        await gather(*tasks)
        self.__start_kafka_consumers_elapsed_time = datetime.now() - start_time

    async def __generate_kafka_consumer_for_subscriber(self, subscriber: MetaSubscriber) -> None:
        executor = await self._create_kafka_consumer_executor(subscriber)
        topics = self.__consume_strategy.get_topics(subscriber)
        kafka_consumer = AIOKafkaConsumer(
            consumer_group=self.__consume_strategy.get_subscription_group(subscriber),
            group_instance_id=self.__generate_consumer_group_instance_id(subscriber),
            topics=topics,
            connection_config=self.__connection_config,
            initial_offset_position=self.__consumer_initial_offset_position,
            partition_assignors=self.__kafka_partition_assignors + self.__FALLBACK_PARTITION_ASSIGNORS,
            logger=self._logger,
            kafka_admin_client=self.__kafka_admin_client,
            auto_create_topic_configuration=self.__auto_create_topic_configuration,
            on_partition_assigned=self.__on_partition_assigned,
            on_partition_revoked=self.__on_partition_revoked,
            session_timeout_ms=self.__session_timeout_ms,
            max_poll_interval_ms=self.__max_poll_interval_ms,
            heartbeat_interval_ms=self.__heartbeat_interval_ms,
            wait_for_connection_to_cluster_ms=self.__wait_for_connection_to_cluster_ms,
        )

        self.__executor_per_consumer_mapper[kafka_consumer] = executor

        self.__queue_per_consumer_mapper[kafka_consumer] = InMemoryMultiqueueRepository()

        self._logger.info(
            f"Initializing consumer group: '{kafka_consumer.get_consumer_group()}' subscribed to the topics: '{kafka_consumer.get_topics()}'"
        )

        await kafka_consumer.init()

        self._logger.info(
            f"Initialized consumer group: '{kafka_consumer.get_consumer_group()}' subscribed to the topics: '{kafka_consumer.get_topics()}'"
        )

    def __generate_consumer_group_instance_id(self, subscriber: MetaSubscriber) -> Optional[str]:
        if self.__worker_instance_id is None:
            return None
        return f"{subscriber.fqn()}-{self.__worker_instance_id}"

    @abstractmethod
    async def _create_kafka_consumer_executor(self, subscriber: MetaSubscriber) -> KafkaEventSubscriberExecutor:
        pass

    async def __polling_task(self) -> None:
        self._logger.info("Creating polling tasks")
        try:
            polling_task_per_consumer = [
                create_task(self.__polling_consuming_tasks(consumer))
                for consumer, subscriber in self.__queue_per_consumer_mapper.items()
            ]

            await gather(*polling_task_per_consumer)

        except Exception:
            self._logger.error(f"Polling task failed with exception: {traceback.format_exc()}")
            self.__should_stop.set()

    async def __polling_consuming_tasks(self, consumer: AIOKafkaConsumer) -> None:
        queue = self.__queue_per_consumer_mapper[consumer]
        while not self.__should_stop.is_set():
            total_size = sum([queue.get_total_size() for queue in self.__queue_per_consumer_mapper.values()])
            if total_size >= self.__max_queue_size:
                await sleep(self.__seconds_between_polls_if_there_are_tasks_in_the_queue)
                continue

            async with self.__polling_tasks_semaphore:
                kafka_poll_records = await consumer.poll(
                    number_of_messages_to_poll=self.__max_records_retrieved_per_poll,
                )

                for kafka_poll_record in kafka_poll_records:
                    queue.push(
                        key=TopicPartition(
                            topic=kafka_poll_record.topic,
                            partition=kafka_poll_record.partition,
                        ),
                        record=kafka_poll_record,
                    )

            if len(kafka_poll_records) == 0:
                await sleep(self.__seconds_between_polls_if_there_are_no_new_tasks)

    async def __consume_events_task(self) -> None:
        self._logger.info("Creating consuming task")
        blocked_tasks_iterator = self.__generate_blocked_consuming_tasks_iterator()

        async for consuming_task in blocked_tasks_iterator:
            consumer = consuming_task.consumer
            kafka_poll_record = consuming_task.kafka_poll_record

            executor = self.__executor_per_consumer_mapper[consumer]
            await executor.consume(kafka_poll_record=kafka_poll_record)
            await consumer.commit_poll_record(kafka_poll_record)

            self.__events_processed += 1

    # This iterator return a blocked task, that will be blocked for other process (like rebalancing), until the next task will be requested
    async def __generate_blocked_consuming_tasks_iterator(self) -> AsyncIterator[ConsumingTask]:
        consumer_queues_cyclic_iterator = cycle(self.__queue_per_consumer_mapper.items())
        last_consumer, _ = next(consumer_queues_cyclic_iterator)

        while not self.__should_stop.is_set():
            if self.is_rebalancing_in_progress():
                self._logger.info("Waiting for rebalancing")
                await sleep(self.__SECONDS_TO_WAIT_BETWEEN_REBALANCING_IN_PROGRESS)
                continue

            if await self.__all_queues_are_empty():
                await sleep(self.__seconds_between_executions_if_there_are_no_tasks_in_the_queue)
                continue

            consumer: Optional[AIOKafkaConsumer] = None

            while consumer != last_consumer:
                consumer, queue = next(consumer_queues_cyclic_iterator)
                async with self.__get_consumer_mutex(
                    consumer_fqn=consumer.get_consumer_group(),
                ):
                    kafka_poll_record = queue.pop()

                    if kafka_poll_record is not None:
                        yield ConsumingTask(consumer, kafka_poll_record)

                    last_consumer = consumer
                    break

    def __get_consumer_mutex(self, consumer_fqn: str) -> Lock:
        mutex_key = f"consumer_{consumer_fqn}"
        return self.__consumer_and_partition_mutex[mutex_key]

    async def __all_queues_are_empty(self) -> bool:
        return all([queue.is_totally_empty() for queue in self.__queue_per_consumer_mapper.values()])

    async def __on_partition_assigned(self, consumer: AIOKafkaConsumer, topics_partitions: set[TopicPartition]) -> None:
        self._logger.info(
            f"rebalancing in progress, assigning partitions {topics_partitions} to consumer {consumer.get_consumer_group()}"
        )
        for topic_partition in topics_partitions:
            self.__queue_per_consumer_mapper[consumer].create_key(topic_partition)

    async def __on_partition_revoked(self, consumer: AIOKafkaConsumer, topics_partitions: set[TopicPartition]) -> None:
        self._logger.info(
            f"rebalancing in progress, revoking partitions {topics_partitions} from consumer {consumer.get_consumer_group()}"
        )

        self.increase_number_of_rebalancing_in_progress()
        async with self.__get_consumer_mutex(
            consumer_fqn=consumer.get_consumer_group(),
        ):
            for topic_partition in topics_partitions:
                self.__queue_per_consumer_mapper[consumer].remove_key(topic_partition)

        self.decrease_number_of_rebalancing_in_progress()

    def increase_number_of_rebalancing_in_progress(self) -> None:
        self.__number_of_rebalancing_processes_in_progress += 1

    def decrease_number_of_rebalancing_in_progress(self) -> None:
        self.__number_of_rebalancing_processes_in_progress -= 1

    def is_rebalancing_in_progress(self) -> bool:
        return self.__number_of_rebalancing_processes_in_progress > 0

    def request_stop(self) -> None:
        self.__should_stop.set()
        self._logger.info("Worker stop requested. Waiting for finalize the current task")

    async def __manage_kafka_consumers_stopping(self) -> None:
        await gather(*[kafka_consumer.stop() for kafka_consumer in self.__queue_per_consumer_mapper.keys()])

    async def __health_check(self, request: web.Request) -> web.Response:
        health_information = {
            "subscribers": [subscriber.fqn() for subscriber in self.__subscribers],
            "number_of_subscribers": len(self.__subscribers),
            "event_processed": self.__events_processed,
        }

        return web.Response(text=json.dumps(health_information), content_type="application/json")

    async def __is_ready(self, request: web.Request) -> web.Response:
        is_ready = self.__is_worked_initialized
        status_code = 200 if is_ready else 503

        self._logger.info(f"Health check is_ready: {is_ready}, status_code: {status_code}")

        return web.Response(
            text=json.dumps({"is_ready": is_ready}), content_type="application/json", status=status_code
        )

    def __print_statistics(self) -> None:
        self._logger.info(
            f"System startup summary:\n"
            f"  - Number of subscribers: {len(self.__subscribers)}\n"
            f"  - Start Kafka consumers elapsed time: {self.__start_kafka_consumers_elapsed_time}\n"
            f"  - Initial coroutines created elapsed time: {self.__initial_coroutines_created_elapsed_time}\n"
            f"  - Events processed: {self.__events_processed}\n"
            f"  - Events processed elapsed time: {self.__events_processed_elapsed_time}"
        )



from typing import Optional

from buz.command import Command
from buz.command.asynchronous import CommandBus
from buz.command.asynchronous import CommandHandler
from buz.command.asynchronous.middleware import HandleMiddleware, HandleMiddlewareChainResolver
from buz.command import MoreThanOneCommandHandlerRelatedException

from buz.locator import Locator


class SelfProcessCommandBus(CommandBus):
    def __init__(
        self,
        locator: Locator[Command, CommandHandler],
        middlewares: Optional[list[HandleMiddleware]] = None,
    ):
        self.__locator = locator
        self.__handle_middleware_chain_resolver = HandleMiddlewareChainResolver(middlewares or [])

    async def handle(self, command: Command) -> None:
        command_handlers = self.__locator.get(command)

        if len(command_handlers) > 1:
            raise MoreThanOneCommandHandlerRelatedException(command, command_handlers)

        command_handler = command_handlers[0]
        await self.__handle_middleware_chain_resolver.resolve(command, command_handler, self.__perform_handle)

    @staticmethod
    async def __perform_handle(command: Command, command_handler: CommandHandler) -> None:
        await command_handler.handle(command)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Optional

from buz.kafka.domain.models.kafka_supported_sasl_mechanisms import KafkaSupportedSaslMechanisms
from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols


@dataclass(frozen=True)
class KafkaConnectionCredentials(ABC):
    security_protocol: KafkaSupportedSecurityProtocols
    user: Optional[str]
    password: Optional[str]
    sasl_mechanism: Optional[KafkaSupportedSaslMechanisms]



from abc import ABC
from typing import Generic, TypeVar

from buz.event.event import Event
from buz.event.subscriber import Subscriber
from buz.event.meta_base_subscriber import MetaBaseSubscriber

TEvent = TypeVar("TEvent", bound=Event)


class BaseSubscriber(Generic[TEvent], Subscriber[TEvent], MetaBaseSubscriber[TEvent], ABC):
    pass



from abc import ABC, abstractmethod


class AsyncConnectionManager(ABC):
    @abstractmethod
    async def connect(self) -> None:
        pass

    @abstractmethod
    async def disconnect(self) -> None:
        pass



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

T = TypeVar("T")


class QueueRepository(ABC, Generic[T]):
    @abstractmethod
    def push(self, record: T):
        pass

    @abstractmethod
    def pop(self) -> T:
        pass

    @abstractmethod
    def get_size(self) -> int:
        pass

    @abstractmethod
    def is_empty(self) -> bool:
        pass



from dataclasses import dataclass
from typing import Optional

from buz.kafka.infrastructure.cdc.cdc_payload import CDCPayload


@dataclass(frozen=True)
class CDCMessage:
    payload: CDCPayload
    schema: Optional[dict]

    def validate(self) -> None:
        if not isinstance(self.payload, CDCPayload):
            raise ValueError("The payload value is not a valid value")
        if self.schema is not None and not isinstance(self.schema, dict):
            raise ValueError("The schema value is not a valid value")

    def __post_init__(self) -> None:
        self.validate()



from __future__ import annotations

from abc import abstractmethod, ABC
from datetime import datetime
from typing import Sequence

from buz.kafka.domain.models.create_kafka_topic import CreateKafkaTopic
from buz.kafka.infrastructure.interfaces.connection_manager import ConnectionManager

DEFAULT_NUMBER_OF_MESSAGES_TO_POLLING = 999


class KafkaAdminClient(ConnectionManager, ABC):
    @abstractmethod
    def create_topics(
        self,
        *,
        topics: Sequence[CreateKafkaTopic],
    ) -> None:
        pass

    @abstractmethod
    def is_topic_created(
        self,
        topic: str,
    ) -> bool:
        pass

    @abstractmethod
    def delete_topics(
        self,
        *,
        topics: set[str],
    ) -> None:
        pass

    @abstractmethod
    def get_topics(
        self,
    ) -> set[str]:
        pass

    @abstractmethod
    def get_number_of_partitions(self, topic: str) -> int:
        pass

    # This function moves the following offset from the provided date
    # if there are no messages with a date greater than the provided offset
    # the offset will be moved to the end
    @abstractmethod
    def move_offsets_to_datetime(
        self,
        *,
        consumer_group: str,
        topic: str,
        target_datetime: datetime,
    ) -> None:
        pass

    @abstractmethod
    def increase_topic_partitions_and_set_offset_of_related_consumer_groups_to_the_beginning_of_the_new_ones(
        self,
        *,
        topic: str,
        new_number_of_partitions: int,
    ) -> None:
        pass



from __future__ import annotations

from typing import Generic, Optional, TypeVar, cast

from kafka import KafkaProducer as KafkaPythonLibraryProducer
from kafka.producer.future import FutureRecordMetadata

from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.models.kafka_supported_compression_type import KafkaSupportedCompressionType
from buz.kafka.domain.services.kafka_producer import KafkaProducer
from buz.kafka.infrastructure.serializers.byte_serializer import ByteSerializer
from buz.kafka.infrastructure.serializers.kafka_header_serializer import KafkaHeaderSerializer

T = TypeVar("T")


class KafkaPythonProducer(KafkaProducer, Generic[T]):
    __kafka_producer: Optional[KafkaPythonLibraryProducer] = None
    __SEND_TIMEOUT_SECONDS = 5

    def __init__(
        self,
        *,
        connection_config: KafkaConnectionConfig,
        byte_serializer: ByteSerializer[T],
        retries: int = 0,
        retry_backoff_ms: int = 100,
        compression_type: Optional[KafkaSupportedCompressionType] = None,
    ) -> None:
        self.__connection_config = connection_config
        self.__byte_serializer = byte_serializer
        self.__header_serializer = KafkaHeaderSerializer()
        self.__retries = retries
        self.__retry_backoff_ms = retry_backoff_ms
        self.__compression_type = compression_type

    def _get_kafka_producer(self) -> KafkaPythonLibraryProducer:
        if self.__kafka_producer is None:
            sasl_mechanism = (
                self.__connection_config.credentials.sasl_mechanism.value
                if self.__connection_config.credentials.sasl_mechanism
                else None
            )
            compression_type = self.__compression_type.value if self.__compression_type else None

            self.__kafka_producer = KafkaPythonLibraryProducer(
                client_id=self.__connection_config.client_id,
                bootstrap_servers=self.__connection_config.bootstrap_servers,
                security_protocol=self.__connection_config.credentials.security_protocol.value,
                sasl_mechanism=sasl_mechanism,
                sasl_plain_username=self.__connection_config.credentials.user,
                sasl_plain_password=self.__connection_config.credentials.password,
                retries=self.__retries,
                retry_backoff_ms=self.__retry_backoff_ms,
                compression_type=compression_type,
            )

        return self.__kafka_producer

    def connect(self):
        self._get_kafka_producer()

    def disconnect(self) -> None:
        if self.__kafka_producer is not None:
            self.__kafka_producer.close()
            self.__kafka_producer = None

    def produce(
        self,
        *,
        topic: str,
        message: T,
        partition_key: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
    ) -> None:
        serialized_headers = self.__header_serializer.serialize(headers) if headers is not None else None
        kafka_producer = self._get_kafka_producer()

        message_future = cast(
            FutureRecordMetadata,
            kafka_producer.send(
                topic=topic,
                value=self.__byte_serializer.serialize(message),
                headers=serialized_headers,
                key=partition_key,
            ),
        )

        # We are forcing a flush because the task related with the send is asynchronous, and we want that the event to be sent after call produce
        message_future.get(self.__SEND_TIMEOUT_SECONDS)



from buz.event.transactional_outbox.outbox_criteria.outbox_sorting_criteria import OutboxSortingCriteria
from buz.event.transactional_outbox.outbox_criteria.outbox_criteria import OutboxCriteria
from buz.event.transactional_outbox.outbox_criteria.outbox_criteria_factory import OutboxCriteriaFactory
from buz.event.transactional_outbox.outbox_criteria.deliverable_records_outbox_criteria_factory import (
    DeliverableRecordsOutboxCriteriaFactory,
)

__all__ = [
    "OutboxSortingCriteria",
    "OutboxCriteria",
    "OutboxCriteriaFactory",
    "DeliverableRecordsOutboxCriteriaFactory",
]



from abc import ABC, abstractmethod
from typing import Awaitable, Generic, Type, TypeVar, Union

from buz import Handler
from buz.event.event import Event

TEvent = TypeVar("TEvent", bound=Event)


class MetaSubscriber(Generic[TEvent], Handler[TEvent], ABC):
    @abstractmethod
    def consume(self, event: TEvent) -> Union[None, Awaitable[None]]:
        pass

    @classmethod
    @abstractmethod
    def handles(cls) -> Type[TEvent]:
        pass



from typing import Sequence
from buz.event import Event, Subscriber
from buz.event.infrastructure.models.execution_context import ExecutionContext
from buz.event.middleware import ConsumeMiddleware, ConsumeCallable
from buz.middleware import MiddlewareChainBuilder


class ConsumeMiddlewareChainResolver:
    def __init__(self, middlewares: Sequence[ConsumeMiddleware]):
        self.__middlewares = middlewares
        self.__middleware_chain_builder: MiddlewareChainBuilder[
            ConsumeCallable, ConsumeMiddleware
        ] = MiddlewareChainBuilder(self.__middlewares)

    def resolve(
        self,
        event: Event,
        subscriber: Subscriber,
        consume: ConsumeCallable,
        execution_context: ExecutionContext,
    ) -> None:
        chain_callable: ConsumeCallable = self.__middleware_chain_builder.get_chain_callable(
            consume, self.__get_middleware_callable
        )
        chain_callable(event, subscriber, execution_context)

    def __get_middleware_callable(
        self, middleware: ConsumeMiddleware, consume_callable: ConsumeCallable
    ) -> ConsumeCallable:
        return lambda event, subscriber, execution_context: middleware.on_consume(
            event, subscriber, consume_callable, execution_context
        )



from abc import abstractmethod

from buz.event import Event
from buz.event.async_subscriber import AsyncSubscriber
from buz.event.infrastructure.models.execution_context import ExecutionContext
from buz.event.middleware.async_consume_middleware import AsyncConsumeMiddleware, AsyncConsumeCallable


class BaseAsyncConsumeMiddleware(AsyncConsumeMiddleware):
    async def on_consume(
        self,
        event: Event,
        subscriber: AsyncSubscriber,
        consume: AsyncConsumeCallable,
        execution_context: ExecutionContext,
    ) -> None:
        await self._before_on_consume(event, subscriber, execution_context)
        await consume(event, subscriber, execution_context)
        await self._after_on_consume(event, subscriber, execution_context)

    @abstractmethod
    async def _before_on_consume(
        self,
        event: Event,
        subscriber: AsyncSubscriber,
        execution_context: ExecutionContext,
    ) -> None:
        pass

    @abstractmethod
    async def _after_on_consume(
        self,
        event: Event,
        subscriber: AsyncSubscriber,
        execution_context: ExecutionContext,
    ) -> None:
        pass



from queue import Queue
from typing import TypeVar

from buz.queue.queue_repository import QueueRepository


T = TypeVar("T")


class InMemoryQueueRepository(QueueRepository[T]):
    def __init__(self, queue: Queue[T]):
        self._queue = queue

    def push(self, record: T):
        self._queue.put(record)

    def pop(self) -> T:
        return self._queue.get()

    def get_size(self) -> int:
        return self._queue.qsize()

    def is_empty(self) -> bool:
        return self._queue.empty()



from __future__ import annotations

from logging import Logger
from typing import Optional

from kafka import KafkaConsumer, KafkaProducer
from kafka.coordinator.assignors.range import RangePartitionAssignor

from buz.kafka.domain.models.kafka_poll_record import KafkaPollRecord
from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition
from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.services.kafka_admin_test_client import (
    DEFAULT_NUMBER_OF_MESSAGES_TO_POLL,
    KafkaAdminTestClient,
)
from buz.kafka.infrastructure.kafka_python.kafka_python_admin_client import KafkaPythonAdminClient
from buz.kafka.infrastructure.kafka_python.translators.consumer_initial_offset_position_translator import (
    KafkaPythonConsumerInitialOffsetPositionTranslator,
)


CONSUMER_POLL_TIMEOUT_MS = 1000


class KafkaPythonAdminTestClient(KafkaPythonAdminClient, KafkaAdminTestClient):
    def __init__(
        self,
        *,
        logger: Logger,
        connection_config: KafkaConnectionConfig,
    ):
        super().__init__(
            connection_config=connection_config,
            logger=logger,
        )

    def send_message_to_topic(
        self,
        *,
        topic: str,
        message: bytes,
        headers: Optional[list[tuple[str, bytes]]] = None,
        partition_key: Optional[str] = None,
    ) -> None:
        producer = KafkaProducer(**self._config_in_library_format)

        producer.send(
            topic=topic,
            value=message,
            headers=headers,
            key=partition_key,
        )
        producer.flush()
        producer.close()

    def get_messages_from_topic(
        self,
        *,
        topic: str,
        consumer_group: Optional[str] = None,
        max_number_of_messages_to_polling: int = DEFAULT_NUMBER_OF_MESSAGES_TO_POLL,
        offset: ConsumerInitialOffsetPosition = ConsumerInitialOffsetPosition.BEGINNING,
    ) -> list[KafkaPollRecord]:
        consumer = KafkaConsumer(
            **self._config_in_library_format,
            group_id=consumer_group,
            enable_auto_commit=True,
            auto_offset_reset=KafkaPythonConsumerInitialOffsetPositionTranslator.to_kafka_supported_format(offset),
            partition_assignment_strategy=(RangePartitionAssignor,),
        )

        consumer.subscribe(topics=[topic])

        records = list(
            consumer.poll(
                timeout_ms=CONSUMER_POLL_TIMEOUT_MS,
                max_records=max_number_of_messages_to_polling,
            ).values()
        )

        consumer.close()

        topic_records = records[0] if len(records) > 0 else []

        return topic_records

    def delete_all_resources(
        self,
    ) -> None:
        self.delete_topics(topics=self.get_topics())
        self.delete_subscription_groups(subscription_groups=self.get_cluster_consumer_groups())
        self._wait_for_cluster_update()



from buz.event.transactional_outbox.outbox_record_finder.outbox_record_stream_finder import OutboxRecordStreamFinder
from buz.event.transactional_outbox.outbox_record_finder.polling_outbox_record_stream_finder import (
    PollingOutboxRecordStreamFinder,
)

__all__ = [
    "OutboxRecordStreamFinder",
    "PollingOutboxRecordStreamFinder",
]



from typing import Optional

from buz.locator import Locator
from buz.query import Query, QueryResponse
from buz.query.synchronous.middleware import HandleMiddleware, HandleMiddlewareChainResolver
from buz.query import MoreThanOneQueryHandlerRelatedException
from buz.query.synchronous import QueryBus, QueryHandler


class SelfProcessQueryBus(QueryBus):
    def __init__(
        self,
        locator: Locator[Query, QueryHandler],
        middlewares: Optional[list[HandleMiddleware]] = None,
    ):
        self.__locator = locator
        self.__handle_middleware_chain_resolver = HandleMiddlewareChainResolver(middlewares or [])

    def handle(self, query: Query) -> QueryResponse:
        query_handlers = self.__locator.get(query)

        if len(query_handlers) > 1:
            raise MoreThanOneQueryHandlerRelatedException(query, query_handlers)

        query_handler = query_handlers[0]
        return self.__handle_middleware_chain_resolver.resolve(query, query_handler, self.__perform_handle)

    @staticmethod
    def __perform_handle(query: Query, query_handler: QueryHandler) -> QueryResponse:
        return query_handler.handle(query)



from itertools import cycle

from buz.event.consumer import Consumer
from buz.event.strategies.execution_strategy.execution_strategy import ExecutionStrategy


class CyclicIteratorExecutionStrategy(ExecutionStrategy):
    def __init__(self, consumers: list[Consumer]):
        self.__consumers: list[Consumer] = consumers
        self.__stop_requested = False

    def start(self) -> None:
        active_consumers: set[Consumer] = set(self.__consumers)

        for consumer in cycle(self.__consumers):
            if consumer not in active_consumers:
                continue
            consumer.run()
            if self.__stop_requested is True:
                active_consumers.remove(consumer)
            if not active_consumers:
                break

    def stop(self) -> None:
        for consumer in self.__consumers:
            consumer.stop()
        self.__stop_requested = True



from dataclasses import dataclass
from buz.kafka.domain.models.kafka_poll_record import KafkaPollRecord
from buz.kafka.infrastructure.aiokafka.aiokafka_consumer import AIOKafkaConsumer


@dataclass(frozen=True)
class ConsumingTask:
    consumer: AIOKafkaConsumer
    kafka_poll_record: KafkaPollRecord



from typing import Type, TypeVar, cast

from buz import Message
from buz.kafka.infrastructure.deserializers.bytes_to_message_deserializer import BytesToMessageDeserializer
from buz.kafka.infrastructure.deserializers.implementations.json_byte_deserializer import JSONByteDeserializer

T = TypeVar("T", bound=Message)


class JSONBytesToMessageDeserializer(BytesToMessageDeserializer[T]):
    def __init__(
        self,
        *,
        event_class: Type[T],
    ):
        self.__event_class = event_class
        self.__json_byte_deserializer = JSONByteDeserializer()

    def deserialize(self, data: bytes) -> T:
        dictionary = self.__json_byte_deserializer.deserialize(data=data)
        return cast(T, self.__event_class.restore(**dictionary))



from __future__ import annotations

from datetime import datetime
from logging import Logger
import re
from typing import Any, Callable, Optional, Sequence, cast

from cachetools import TTLCache
from kafka import KafkaClient, KafkaConsumer
from kafka.admin import KafkaAdminClient as KafkaPythonLibraryAdminClient, NewTopic
from kafka.admin.new_partitions import NewPartitions
from kafka.errors import TopicAlreadyExistsError
from kafka.structs import TopicPartition, OffsetAndTimestamp

from buz.kafka.domain.exceptions.not_all_partition_assigned_exception import NotAllPartitionAssignedException
from buz.kafka.domain.exceptions.not_valid_partition_number_exception import NotValidPartitionNumberException
from buz.kafka.domain.exceptions.topic_already_created_exception import KafkaTopicsAlreadyCreatedException
from buz.kafka.domain.exceptions.topic_not_found_exception import TopicNotFoundException
from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition
from buz.kafka.domain.models.create_kafka_topic import CreateKafkaTopic
from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.services.kafka_admin_client import KafkaAdminClient

from buz.kafka.infrastructure.kafka_python.translators.consumer_initial_offset_position_translator import (
    KafkaPythonConsumerInitialOffsetPositionTranslator,
)

INTERNAL_KAFKA_TOPICS = {"__consumer_offsets", "_schema"}
TOPIC_CACHE_KEY = "topics"


class KafkaPythonAdminClient(KafkaAdminClient):
    __PYTHON_KAFKA_DUPLICATED_TOPIC_ERROR_CODE = 36

    _kafka_admin: Optional[KafkaPythonLibraryAdminClient] = None
    _kafka_client: Optional[KafkaClient] = None

    def __init__(
        self,
        *,
        logger: Logger,
        connection_config: KafkaConnectionConfig,
        cache_ttl_seconds: int = 0,
    ):
        self._logger = logger
        self.__connection_config = connection_config
        self._config_in_library_format = self.__get_kafka_config_in_library_format(self.__connection_config)
        self.__ttl_cache: TTLCache[str, Any] = TTLCache(maxsize=1, ttl=cache_ttl_seconds)

    def __get_kafka_config_in_library_format(self, config: KafkaConnectionConfig) -> dict:
        return {
            "client_id": config.client_id,
            "bootstrap_servers": config.bootstrap_servers,
            "security_protocol": config.credentials.security_protocol.value,
            "sasl_mechanism": config.credentials.sasl_mechanism.value if config.credentials.sasl_mechanism else None,
            "sasl_plain_username": config.credentials.user,
            "sasl_plain_password": config.credentials.password,
        }

    def connect(self):
        self._get_kafka_admin()
        self._get_kafka_client()

    def disconnect(self):
        if self._kafka_admin is not None:
            self._kafka_admin.close()
            self._kafka_admin = None
        if self._kafka_client is not None:
            self._kafka_client.close()
            self._kafka_client = None

    def _get_kafka_admin(self) -> KafkaPythonLibraryAdminClient:
        if not self._kafka_admin:
            self._kafka_admin = KafkaPythonLibraryAdminClient(**self._config_in_library_format)
        return self._kafka_admin

    def _get_kafka_client(self) -> KafkaClient:
        if not self._kafka_client:
            self._kafka_client = KafkaClient(**self._config_in_library_format)
        return self._kafka_client

    def create_topics(
        self,
        *,
        topics: Sequence[CreateKafkaTopic],
    ) -> None:
        new_topics = [
            NewTopic(
                name=topic.name,
                num_partitions=topic.partitions,
                replication_factor=topic.replication_factor,
                topic_configs=topic.configs,
            )
            for topic in topics
        ]

        try:
            self._get_kafka_admin().create_topics(new_topics=new_topics)
        except TopicAlreadyExistsError as error:
            topic_names = self.__get_list_of_kafka_topics_from_topic_already_exists_error(error)
            raise KafkaTopicsAlreadyCreatedException(topic_names=topic_names)

    def __get_list_of_kafka_topics_from_topic_already_exists_error(self, error: TopicAlreadyExistsError) -> list[str]:
        message = str(error)
        response_message = re.search(r"topic_errors=\[.*?]", message)
        topic_messages = re.findall(
            r"topic='[^']*', error_code=" + str(self.__PYTHON_KAFKA_DUPLICATED_TOPIC_ERROR_CODE), response_message[0]  # type: ignore
        )

        return [re.search("'.*'", topic_message)[0].strip("'") for topic_message in topic_messages]  # type: ignore

    def is_topic_created(
        self,
        topic: str,
    ) -> bool:
        return topic in self.get_topics()

    def get_topics(
        self,
    ) -> set[str]:
        return self.__resolve_cached_property(
            TOPIC_CACHE_KEY, lambda: set(self._get_kafka_admin().list_topics()) - INTERNAL_KAFKA_TOPICS
        )

    def __resolve_cached_property(self, property_key: str, callback: Callable) -> Any:
        value = self.__ttl_cache.get(property_key)
        if value is not None:
            return value
        value = callback()
        self.__ttl_cache[property_key] = value
        return value

    def delete_topics(
        self,
        *,
        topics: set[str],
    ) -> None:
        self._get_kafka_admin().delete_topics(
            topics=topics,
        )
        self.__remove_cache_property(TOPIC_CACHE_KEY)

    def __remove_cache_property(self, property_key: str) -> None:
        self.__ttl_cache.pop(property_key, None)

    def delete_subscription_groups(
        self,
        *,
        subscription_groups: set[str],
    ) -> None:
        self._get_kafka_admin().delete_consumer_groups(
            group_ids=subscription_groups,
        )

    def get_cluster_consumer_groups(
        self,
    ) -> set[str]:
        return set([consumer_group_tuple[0] for consumer_group_tuple in self._get_kafka_admin().list_consumer_groups()])

    def _wait_for_cluster_update(self) -> None:
        future = self._get_kafka_client().cluster.request_update()
        self._get_kafka_client().poll(future=future)

    def move_offsets_to_datetime(
        self,
        *,
        consumer_group: str,
        topic: str,
        target_datetime: datetime,
    ) -> None:
        (consumer, topic_partitions) = self.__get_consumer_with_all_partitions_assigned(
            consumer_group=consumer_group,
            topic=topic,
        )

        offsets_for_date = self.__get_first_offset_after_date(
            consumer=consumer,
            topic_partitions=topic_partitions,
            target_datetime=target_datetime,
        )

        try:
            end_offsets = consumer.end_offsets(topic_partitions)

            if end_offsets is None or len(end_offsets.keys()) != len(topic_partitions):
                raise Exception(f'There was an error extracting the end offsets of the topic "{topic}"')

            for topic_partition in topic_partitions:
                offset_and_timestamp = offsets_for_date.get(topic_partition)
                if offset_and_timestamp:
                    self._logger.info(f'moving "{topic_partition}" to the offset "{offset_and_timestamp.offset}"')
                    consumer.seek(topic_partition, offset_and_timestamp.offset)
                else:
                    self._logger.info(
                        f'moving "{topic_partition}" to the end of the topic because there are no messages later than "{target_datetime}"'
                    )
                    consumer.seek(topic_partition, end_offsets[topic_partition])

            consumer.commit()
        except Exception as exception:
            consumer.close()
            raise exception

        consumer.close()

    def __get_consumer_with_all_partitions_assigned(
        self,
        consumer_group: str,
        topic: str,
    ) -> tuple[KafkaConsumer, Sequence[TopicPartition]]:
        consumer = KafkaConsumer(
            group_id=consumer_group,
            enable_auto_commit=False,
            auto_offset_reset=KafkaPythonConsumerInitialOffsetPositionTranslator.to_kafka_supported_format(
                ConsumerInitialOffsetPosition.BEGINNING
            ),
            **self._config_in_library_format,
        )

        try:
            partitions = self.get_number_of_partitions(topic)

            topic_partitions = [TopicPartition(topic=topic, partition=partition) for partition in range(partitions)]

            consumer.subscribe(topic)

            self.__force_partition_assignment(consumer)

            # We need all the partitions in order to update the offsets
            if len(consumer.assignment()) != len(topic_partitions):
                raise NotAllPartitionAssignedException(
                    topic_name=topic,
                    consumer_group=consumer_group,
                )

            # This could produce a race condition, but it is a limitation of kafka admin (we are not able to check if all the partition are assigned using the manual assignment)
            # https://github.com/dpkp/kafka-python/blob/master/kafka/consumer/group.py#L430
            consumer.unsubscribe()
            consumer.assign(topic_partitions)
            self.__force_partition_assignment(consumer)

            return (consumer, topic_partitions)
        except Exception as exception:
            consumer.close()
            raise exception

    def __get_first_offset_after_date(
        self,
        *,
        consumer: KafkaConsumer,
        topic_partitions: Sequence[TopicPartition],
        target_datetime: datetime,
    ) -> dict[TopicPartition, Optional[OffsetAndTimestamp]]:
        offset_for_times: dict[TopicPartition, Optional[int]] = {}
        timestamp_ms = int(target_datetime.timestamp() * 1000)

        for topic_partition in topic_partitions:
            offset_for_times[topic_partition] = timestamp_ms

        return cast(
            dict[TopicPartition, Optional[OffsetAndTimestamp]],
            consumer.offsets_for_times(offset_for_times),
        )

    # We are not to commit the new offset, but we need to execute a polling in order to start the partition assignment
    def __force_partition_assignment(self, consumer: KafkaConsumer) -> None:
        consumer.poll(max_records=1, timeout_ms=0)

    def increase_topic_partitions_and_set_offset_of_related_consumer_groups_to_the_beginning_of_the_new_ones(
        self,
        *,
        topic: str,
        new_number_of_partitions: int,
    ) -> None:
        self._logger.info(
            f'Increasing topic "{topic}" partitions: Verifying the new number of partitions "{new_number_of_partitions}"'
        )

        previous_partitions_number = self.get_number_of_partitions(topic)
        topic_partitions = [
            TopicPartition(topic=topic, partition=partition) for partition in range(previous_partitions_number)
        ]

        if previous_partitions_number >= new_number_of_partitions:
            raise NotValidPartitionNumberException(
                partition_number=new_number_of_partitions,
                min_partition_number=len(topic_partitions),
            )

        self._logger.info(f'Increasing topic "{topic}" partitions: Extracting related consumer groups')
        related_consumer_groups = self.__get_consumer_groups_related_to_a_topic(topic_partitions)

        self._logger.info(
            f'Increasing topic "{topic}" partitions: The following consumer groups will be updated:"{related_consumer_groups}"'
        )

        consumers_to_update: list[KafkaConsumer] = []
        new_partitions_consumer: Optional[KafkaConsumer] = None

        try:
            for consumer_group in related_consumer_groups:
                self._logger.info(
                    f'Increasing topic "{topic}" partitions: Requesting the assignment of the partitions of the group "{consumer_group}"'
                )
                (consumer_with_all_partitions, _) = self.__get_consumer_with_all_partitions_assigned(
                    consumer_group=consumer_group,
                    topic=topic,
                )
                consumers_to_update.append(consumer_with_all_partitions)

            self._logger.info(
                f'Increasing topic "{topic}" partitions: Incrementing the partition to "{new_number_of_partitions}"'
            )

            self._get_kafka_admin().create_partitions(
                {
                    topic: NewPartitions(total_count=new_number_of_partitions),
                }
            )

            new_partitions = [
                TopicPartition(
                    topic=topic,
                    partition=partition_index,
                )
                for partition_index in range(previous_partitions_number, new_number_of_partitions)
            ]

            for consumer_group in related_consumer_groups:
                self._logger.info(
                    f'Increasing topic "{topic}" partitions: Moving the offset of the consumer group "{consumer_group}" to the beginning of the new partitions'
                )
                # We need to create a new consumer because kafka-python has a limitation that does not allow to assign specific partitions to a consumer subscribed to an entire topic
                new_partitions_consumer = KafkaConsumer(
                    group_id=consumer_group,
                    enable_auto_commit=False,
                    auto_offset_reset=KafkaPythonConsumerInitialOffsetPositionTranslator.to_kafka_supported_format(
                        ConsumerInitialOffsetPosition.BEGINNING
                    ),
                    **self._config_in_library_format,
                )
                new_partitions_consumer.assign(new_partitions)
                for new_partition in new_partitions:
                    new_partitions_consumer.seek(new_partition, 0)
                new_partitions_consumer.commit()
                new_partitions_consumer.close()

            self._logger.info(f'Increasing topic "{topic}" partitions: Process complete')

        except Exception as exception:
            for consumer_with_all_partitions in consumers_to_update:
                consumer_with_all_partitions.close()

            if new_partitions_consumer is not None:
                new_partitions_consumer.close()

            self._logger.error(f'Increasing topic "{topic}" partitions: unexpected error {exception}')
            raise exception

        return

    def get_number_of_partitions(self, topic: str) -> int:
        consumer = KafkaConsumer(
            enable_auto_commit=False,
            auto_offset_reset=KafkaPythonConsumerInitialOffsetPositionTranslator.to_kafka_supported_format(
                ConsumerInitialOffsetPosition.BEGINNING
            ),
            **self._config_in_library_format,
        )

        try:
            partitions = consumer.partitions_for_topic(topic)
            if partitions is None:
                raise TopicNotFoundException(topic_name=topic)

            return len(partitions)
        except Exception as exception:
            consumer.close()
            raise exception

    # The purpose of this function is to get all the consumer groups that are consuming from the topic
    # It is a heavy tasks because we need to get the offset of all the partitions of the topic
    def __get_consumer_groups_related_to_a_topic(self, topic_partitions: Sequence[TopicPartition]) -> set[str]:
        cluster_consumer_groups = self.get_cluster_consumer_groups()

        related_consumer_groups: set[str] = set()

        for consumer_group in cluster_consumer_groups:
            partitions_offsets = list(
                self._get_kafka_admin()
                .list_consumer_group_offsets(consumer_group, partitions=topic_partitions)
                .values()
            )

            partitions_with_valid_offsets = [partition for partition in partitions_offsets if partition.offset != -1]

            if len(partitions_with_valid_offsets) == 0:
                continue

            related_consumer_groups.add(consumer_group)

        return related_consumer_groups



from dataclasses import asdict
from typing import Type, TypeVar, cast

from buz.event import Event
from buz.kafka.infrastructure.serializers.implementations.json_byte_serializer import (
    JSON_SERIALIZABLE,
    JSONByteSerializer,
)
from buz.serializer.message_to_bytes_serializer import MessageToByteSerializer

T = TypeVar("T", bound=Event)


class MessageClassToJSONByteSerializer(MessageToByteSerializer[T]):
    def __init__(
        self,
        *,
        event_class: Type[T],
    ):
        self.__event_class = event_class
        self.__json_byte_serializer = JSONByteSerializer()

    def serialize(self, data: T) -> bytes:
        dictionary = cast(JSON_SERIALIZABLE, asdict(data))
        return self.__json_byte_serializer.serialize(dictionary)



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

from buz.kafka.domain.models.kafka_connection_credentials import KafkaConnectionCredentials
from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols


@dataclass(frozen=True)
class KafkaConnectionPlainTextCredentials(KafkaConnectionCredentials):
    security_protocol: Literal[KafkaSupportedSecurityProtocols.PLAINTEXT] = KafkaSupportedSecurityProtocols.PLAINTEXT
    user: None = None
    password: None = None
    sasl_mechanism: None = None



from abc import abstractmethod
from typing import Callable

from buz.command import Command
from buz.command.synchronous import CommandHandler
from buz.middleware import Middleware

HandleCallable = Callable[[Command, CommandHandler], None]


class HandleMiddleware(Middleware):
    @abstractmethod
    def on_handle(self, command: Command, command_handler: CommandHandler, handle: HandleCallable) -> None:
        pass



from logging import Logger
from typing import Optional, Iterable

from buz.event import Event, EventBus
from buz.event.exceptions.event_not_published_exception import EventNotPublishedException
from buz.event.infrastructure.buz_kafka.exceptions.kafka_event_bus_config_not_valid_exception import (
    KafkaEventBusConfigNotValidException,
)
from buz.event.infrastructure.buz_kafka.publish_strategy.publish_strategy import KafkaPublishStrategy
from buz.event.middleware import (
    PublishMiddleware,
)
from buz.event.middleware.publish_middleware_chain_resolver import PublishMiddlewareChainResolver
from buz.kafka.domain.exceptions.topic_already_created_exception import KafkaTopicsAlreadyCreatedException
from buz.kafka.domain.models.auto_create_topic_configuration import AutoCreateTopicConfiguration
from buz.kafka.domain.models.create_kafka_topic import CreateKafkaTopic
from buz.kafka.domain.services.kafka_admin_client import KafkaAdminClient
from buz.kafka.domain.services.kafka_producer import KafkaProducer


class BuzKafkaEventBus(EventBus):
    def __init__(
        self,
        *,
        publish_strategy: KafkaPublishStrategy,
        producer: KafkaProducer,
        logger: Logger,
        kafka_admin_client: Optional[KafkaAdminClient] = None,
        publish_middlewares: Optional[list[PublishMiddleware]] = None,
        auto_create_topic_configuration: Optional[AutoCreateTopicConfiguration] = None,
    ):
        self.__publish_middleware_chain_resolver = PublishMiddlewareChainResolver(publish_middlewares or [])
        self.__publish_strategy = publish_strategy
        self.__producer = producer
        self.__topics_checked: dict[str, bool] = {}
        self.__kafka_admin_client = kafka_admin_client
        self.__auto_create_topic_configuration = auto_create_topic_configuration
        self.__logger = logger
        self.__check_kafka_admin_client_is_needed()

    def __check_kafka_admin_client_is_needed(self) -> None:
        if self.__kafka_admin_client is None and self.__auto_create_topic_configuration is not None:
            raise KafkaEventBusConfigNotValidException(
                "A KafkaAdminClient is needed to create topics when 'auto_create_topic_configuration' is set."
            )

    def publish(self, event: Event) -> None:
        self.__publish_middleware_chain_resolver.resolve(event, self.__perform_publish)

    def __perform_publish(self, event: Event) -> None:
        try:
            topic = self.__publish_strategy.get_topic(event)

            if self.__auto_create_topic_configuration is not None and self.__is_topic_created(topic) is False:
                try:
                    self.__logger.info(f"Creating missing topic: {topic}..")
                    self.__get_kafka_admin_client().create_topics(
                        topics=[
                            CreateKafkaTopic(
                                name=topic,
                                partitions=self.__auto_create_topic_configuration.partitions,
                                replication_factor=self.__auto_create_topic_configuration.replication_factor,
                                configs=self.__auto_create_topic_configuration.configs,
                            )
                        ]
                    )
                    self.__logger.info(f"Created missing topic: {topic}")
                    self.__topics_checked[topic] = True
                except KafkaTopicsAlreadyCreatedException:
                    pass

            headers = self.__get_event_headers(event)
            self.__producer.produce(
                message=event,
                headers=headers,
                topic=topic,
            )
        except Exception as exc:
            raise EventNotPublishedException(event) from exc

    def __get_kafka_admin_client(self) -> KafkaAdminClient:
        if self.__kafka_admin_client is None:
            raise KafkaEventBusConfigNotValidException("KafkaAdminClient is not set.")
        return self.__kafka_admin_client

    def __is_topic_created(self, topic: str) -> bool:
        is_topic_created = self.__topics_checked.get(topic, None)

        if is_topic_created is not None:
            return is_topic_created

        is_topic_created = self.__get_kafka_admin_client().is_topic_created(topic)
        self.__topics_checked[topic] = is_topic_created

        return is_topic_created

    def bulk_publish(self, events: Iterable[Event]) -> None:
        for event in events:
            self.publish(event)

    def __get_event_headers(self, event: Event) -> dict:
        return {"id": event.id}



from __future__ import annotations


class TopicNotFoundException(Exception):
    def __init__(self, topic_name: str) -> None:
        super().__init__(f'The topic "{topic_name}", has not been found')



from asyncio import get_running_loop
from logging import Logger
import time
from typing import Optional, Sequence, cast
from buz.event import Event
from buz.event.infrastructure.buz_kafka.consume_strategy.kafka_on_fail_strategy import KafkaOnFailStrategy
from buz.event.infrastructure.buz_kafka.exceptions.max_consumer_retry_exception import MaxConsumerRetryException
from buz.event.infrastructure.buz_kafka.exceptions.retry_exception import ConsumerRetryException
from buz.event.infrastructure.buz_kafka.kafka_event_subscriber_executor import KafkaEventSubscriberExecutor
from buz.event.infrastructure.buz_kafka.models.kafka_delivery_context import KafkaDeliveryContext
from buz.event.infrastructure.models.execution_context import ExecutionContext
from buz.event.middleware.consume_middleware import ConsumeMiddleware
from buz.event.middleware.consume_middleware_chain_resolver import ConsumeMiddlewareChainResolver
from buz.event.strategies.retry.consume_retrier import ConsumeRetrier
from buz.event.strategies.retry.reject_callback import RejectCallback
from buz.event.subscriber import Subscriber
from buz.kafka.domain.exceptions.not_valid_kafka_message_exception import NotValidKafkaMessageException
from buz.kafka.domain.models.kafka_consumer_record import KafkaConsumerRecord
from buz.kafka.domain.models.kafka_poll_record import KafkaPollRecord
from buz.kafka.infrastructure.deserializers.byte_deserializer import ByteDeserializer
from buz.kafka.infrastructure.serializers.kafka_header_serializer import KafkaHeaderSerializer


class KafkaEventSyncSubscriberExecutor(KafkaEventSubscriberExecutor):
    def __init__(
        self,
        *,
        subscriber: Subscriber,
        logger: Logger,
        consume_middlewares: Optional[Sequence[ConsumeMiddleware]] = None,
        seconds_between_retries: float = 5,
        byte_deserializer: ByteDeserializer[Event],
        header_deserializer: KafkaHeaderSerializer,
        on_fail_strategy: KafkaOnFailStrategy,
        consume_retrier: Optional[ConsumeRetrier] = None,
        reject_callback: Optional[RejectCallback] = None,
    ):
        self.__subscriber = subscriber
        self.__logger = logger
        self.__consume_middleware_chain_resolver = ConsumeMiddlewareChainResolver(consume_middlewares or [])
        self.__seconds_between_retires = seconds_between_retries
        self.__on_fail_strategy = on_fail_strategy
        self.__consume_retrier = consume_retrier
        self.__reject_callback = reject_callback
        self.__byte_deserializer = byte_deserializer
        self.__header_deserializer = header_deserializer

    async def consume(
        self,
        *,
        kafka_poll_record: KafkaPollRecord,
    ) -> None:
        try:
            if kafka_poll_record.value is None:
                raise NotValidKafkaMessageException("Message is None")

            kafka_record_value = cast(bytes, kafka_poll_record.value)

            deserialized_value = self.__byte_deserializer.deserialize(kafka_record_value)

            self.__logger.info(
                f"consuming the event '{deserialized_value.id}' by the subscriber '{self.__subscriber.fqn()}', "
                + f"topic: '{kafka_poll_record.topic}', partition: '{kafka_poll_record.partition}', offset: '{kafka_poll_record.offset}'"
            )

            await get_running_loop().run_in_executor(
                None,
                lambda: self.__execution_callback(
                    subscriber=self.__subscriber,
                    message=KafkaConsumerRecord(
                        value=deserialized_value,
                        headers=self.__header_deserializer.deserialize(kafka_poll_record.headers),
                    ),
                    execution_context=ExecutionContext(
                        delivery_context=KafkaDeliveryContext(
                            topic=kafka_poll_record.topic,
                            consumer_group=self.__subscriber.fqn(),
                            partition=kafka_poll_record.partition,
                            timestamp=kafka_poll_record.timestamp,
                        )
                    ),
                ),
            )

        except NotValidKafkaMessageException:
            self.__logger.error(
                f'The message "{str(kafka_poll_record.value)}" is not valid, it will be consumed but not processed'
            )

    def __execution_callback(
        self, subscriber: Subscriber, message: KafkaConsumerRecord[Event], execution_context: ExecutionContext
    ) -> None:
        self.__consume_middleware_chain_resolver.resolve(
            event=message.value,
            subscriber=subscriber,
            execution_context=execution_context,
            consume=self.__perform_consume,
        )

    def __perform_consume(self, event: Event, subscriber: Subscriber, execution_context: ExecutionContext) -> None:
        number_of_executions = 0
        should_retry = True

        if self.__consume_retrier is not None:
            should_retry = self.__consume_retrier.should_retry(event, [subscriber])

        if should_retry is False:
            max_retry_exception = MaxConsumerRetryException(
                event_id=event.id,
                subscriber_fqn=subscriber.fqn(),
            )

            if self.__on_fail_strategy == KafkaOnFailStrategy.STOP_ON_FAIL:
                raise max_retry_exception

            self.__logger.warning(
                f"The event {event.id} with the subscriber {subscriber.fqn()} has reach the max number of retries, it will be consumed but not processed"
            )

            if self.__reject_callback is not None:
                self.__reject_callback.on_reject(event=event, subscribers=[subscriber], exception=max_retry_exception)

            return

        while should_retry is True:
            try:
                self.__register_retry(event, subscriber)
                number_of_executions += 1
                subscriber.consume(event)
                self.__clean_retries(event, subscriber)
                return
            except Exception as exception:
                if self.__should_retry(event, subscriber) is True:
                    self.__logger.warning(
                        ConsumerRetryException(
                            number_of_executions=number_of_executions,
                            event_id=event.id,
                            subscriber_fqn=subscriber.fqn(),
                        ),
                        exc_info=exception,
                    )
                    time.sleep(self.__seconds_between_retires)
                    continue

                self.__logger.exception(exception)

                if self.__reject_callback is not None:
                    self.__reject_callback.on_reject(event=event, subscribers=[subscriber], exception=exception)

                if self.__on_fail_strategy == KafkaOnFailStrategy.STOP_ON_FAIL:
                    raise exception

                return

    def __should_retry(self, event: Event, subscriber: Subscriber) -> bool:
        if self.__consume_retrier is None:
            return False

        return self.__consume_retrier.should_retry(event, [subscriber])

    def __register_retry(self, event: Event, subscriber: Subscriber) -> None:
        if self.__consume_retrier is None:
            return None

        return self.__consume_retrier.register_retry(event, [subscriber])

    def __clean_retries(self, event: Event, subscriber: Subscriber) -> None:
        if self.__consume_retrier is None:
            return None

        return self.__consume_retrier.clean_retries(event, [subscriber])



from abc import abstractmethod
from typing import Awaitable, Callable

from buz.event import Event
from buz.event.async_subscriber import AsyncSubscriber
from buz.event.infrastructure.models.execution_context import ExecutionContext
from buz.middleware import Middleware

AsyncConsumeCallable = Callable[[Event, AsyncSubscriber, ExecutionContext], Awaitable[None]]


class AsyncConsumeMiddleware(Middleware):
    @abstractmethod
    async def on_consume(
        self,
        event: Event,
        subscriber: AsyncSubscriber,
        consume: AsyncConsumeCallable,
        execution_context: ExecutionContext,
    ) -> None:
        pass



from buz.event import Event
from buz.event.middleware import PublishMiddleware, PublishCallable
from buz.middleware import MiddlewareChainBuilder


class PublishMiddlewareChainResolver:
    def __init__(self, middlewares: list[PublishMiddleware]):
        self.__middlewares = middlewares
        self.__middleware_chain_builder: MiddlewareChainBuilder[
            PublishCallable, PublishMiddleware
        ] = MiddlewareChainBuilder(middlewares)

    def resolve(self, event: Event, publish: PublishCallable) -> None:
        chain_callable: PublishCallable = self.__middleware_chain_builder.get_chain_callable(
            publish, self.__get_middleware_callable
        )
        chain_callable(event)

    def __get_middleware_callable(
        self, middleware: PublishMiddleware, publish_callable: PublishCallable
    ) -> PublishCallable:
        return lambda event: middleware.on_publish(event, publish_callable)



from collections import defaultdict
from typing import DefaultDict, Generic, TypeVar, Type, cast

from buz import Handler
from buz import Message
from buz.locator import Locator, HandlerFqnNotFoundException, MessageFqnNotFoundException
from buz.locator.sync import HandlerAlreadyRegisteredException
from buz.locator.sync import HandlerNotRegisteredException

K = TypeVar("K", bound=Message)
V = TypeVar("V", bound=Handler)


class InstanceLocator(Locator, Generic[K, V]):
    def __init__(self) -> None:
        self.__mapping: DefaultDict[str, list[V]] = defaultdict(list)

    def register(self, handler: V) -> None:
        message_fqn = handler.handles().fqn()
        self.__guard_handler_already_registered(message_fqn, handler)
        self.__mapping[message_fqn].append(handler)

    def __guard_handler_already_registered(self, message_fqn: str, handler: V) -> None:
        if handler in self.__mapping[message_fqn]:
            raise HandlerAlreadyRegisteredException(handler)

    def unregister(self, handler: V) -> None:
        message_fqn = handler.handles().fqn()
        self.__guard_subscriber_not_registered(message_fqn, handler)
        self.__mapping[message_fqn].remove(handler)

    def __guard_subscriber_not_registered(self, message_fqn: str, handler: V) -> None:
        if handler not in self.__mapping[message_fqn]:
            raise HandlerNotRegisteredException(handler)

    def get(self, message: K) -> list[V]:
        return self.__mapping.get(message.fqn(), [])

    def get_handler_by_fqn(self, handler_fqn: str) -> V:
        for handlers in self.__mapping.values():
            for handler in handlers:
                if handler.fqn() == handler_fqn:
                    return handler
        raise HandlerFqnNotFoundException(handler_fqn)

    def get_message_klass_by_fqn(self, message_fqn: str) -> Type[K]:
        try:
            handler = self.__mapping.get(message_fqn, [])[0]
            return cast(Type[K], handler.handles())
        except (IndexError, TypeError):
            raise MessageFqnNotFoundException(message_fqn)



from abc import ABC, abstractmethod

from buz.event.transactional_outbox.outbox_criteria import OutboxCriteria


class OutboxCriteriaFactory(ABC):
    @abstractmethod
    def create(self) -> OutboxCriteria:
        pass



from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar

from buz import Handler
from buz.query import Query, QueryResponse

TQuery = TypeVar("TQuery", bound=Query)
TQueryResponse = TypeVar("TQueryResponse", bound=QueryResponse)


class QueryHandler(Generic[TQuery, TQueryResponse], Handler[TQuery], ABC):
    @classmethod
    @abstractmethod
    def handles(cls) -> Type[TQuery]:
        pass

    @abstractmethod
    async def handle(self, query: TQuery) -> TQueryResponse:
        pass



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from buz.event.event import Event
from buz.event.meta_subscriber import MetaSubscriber

TEvent = TypeVar("TEvent", bound=Event)


class Subscriber(Generic[TEvent], MetaSubscriber[TEvent], ABC):
    @abstractmethod
    def consume(self, event: TEvent) -> None:
        pass



from buz.event.middleware.base_async_consume_middleware import BaseAsyncConsumeMiddleware
from buz.event.middleware.publish_middleware import PublishMiddleware, PublishCallable
from buz.event.middleware.base_publish_middleware import BasePublishMiddleware
from buz.event.middleware.publish_middleware_chain_resolver import PublishMiddlewareChainResolver
from buz.event.middleware.consume_middleware import ConsumeMiddleware, ConsumeCallable
from buz.event.middleware.base_consume_middleware import BaseConsumeMiddleware
from buz.event.middleware.consume_middleware_chain_resolver import ConsumeMiddlewareChainResolver

__all__ = [
    "PublishMiddleware",
    "PublishCallable",
    "BasePublishMiddleware",
    "PublishMiddlewareChainResolver",
    "ConsumeMiddleware",
    "ConsumeCallable",
    "BaseConsumeMiddleware",
    "ConsumeMiddlewareChainResolver",
    "BaseAsyncConsumeMiddleware",
]



from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar

from buz import Message


TMessage = TypeVar("TMessage", bound=Message)


class Handler(Generic[TMessage], ABC):
    @classmethod
    @abstractmethod
    def handles(cls) -> Type[TMessage]:
        pass

    @classmethod
    @abstractmethod
    def fqn(cls) -> str:
        pass



from abc import ABC, abstractmethod
from typing import Collection

from buz.event import Event
from buz.kafka.infrastructure.interfaces.async_connection_manager import AsyncConnectionManager


class AsyncEventBus(AsyncConnectionManager, ABC):
    @abstractmethod
    async def publish(self, event: Event) -> None:
        pass

    @abstractmethod
    async def bulk_publish(self, events: Collection[Event]) -> None:
        pass



from datetime import datetime
from logging import Logger
from typing import Awaitable, Callable, Optional

from aiokafka import ConsumerRebalanceListener, TopicPartition


# @see = https://aiokafka.readthedocs.io/en/stable/api.html#aiokafka.abc.ConsumerRebalanceListener.on_partitions_assigned
class KafkaCallbackRebalancer(ConsumerRebalanceListener):
    def __init__(
        self,
        logger: Logger,
        on_partitions_revoked: Callable[[set[TopicPartition]], Awaitable[None]],
        on_partitions_assigned: Callable[[set[TopicPartition]], Awaitable[None]],
    ):
        self.__logger = logger
        self.__on_partition_revoked = on_partitions_revoked
        self.__on_partition_assigned = on_partitions_assigned
        self.__assigned_partitions: set[TopicPartition] = set()
        self.__rebalancing_start_time: Optional[datetime] = None

    async def on_partitions_revoked(self, revoked: set[TopicPartition]) -> None:
        self.__rebalancing_start_time = datetime.now()

        if len(revoked) == 0:
            return None

        self.__logger.info(f"Partitions revoked by the rebalancing process: '{revoked}'")

        await self.__on_partition_revoked(revoked)
        self.__assigned_partitions.difference_update(revoked)

        self.__logger.info(f"Partitions after revoking process: '{self.__assigned_partitions}'")

    async def on_partitions_assigned(self, assigned: set[TopicPartition]) -> None:
        new_partitions_assigned = assigned.difference(self.__assigned_partitions)

        if len(new_partitions_assigned) == 0:
            return None

        self.__logger.info(f"Partitions assigned by the rebalancing process: '{new_partitions_assigned}'")

        await self.__on_partition_assigned(new_partitions_assigned)
        self.__assigned_partitions.update(new_partitions_assigned)

        self.__logger.info(f"Partitions after assigning process: '{self.__assigned_partitions}'")

        if self.__rebalancing_start_time is not None:
            elapsed_time_ms = int((datetime.now() - self.__rebalancing_start_time).total_seconds() * 1000)
            self.__logger.info(
                f"Rebalancing for topic {set(v.topic for v in assigned)} ended, elapsed time: {elapsed_time_ms} ms"
            )



class HandlerNotFoundException(Exception):
    def __init__(self, handler_id: str):
        self.handler_id = handler_id
        super().__init__(f"Handler with id {handler_id} has not been found in the container and cannot be registered")



from queue import Queue, Empty
from typing import Optional, TypeVar, cast

from buz.queue.multiqueue_repository import MultiqueueRepository

K = TypeVar("K")
R = TypeVar("R")


class InMemoryMultiqueueRepository(MultiqueueRepository[K, R]):
    def __init__(self):
        self.__queues = cast(dict[K, Queue[R]], {})
        self.__last_key_index = 0

    def create_key(self, key: K) -> None:
        self.__queues[key] = Queue[R]()

    def remove_key(self, key: K) -> None:
        if key not in self.__queues:
            return

        self.__queues.pop(key)

    def push(self, key: K, record: R) -> None:
        if key not in self.__queues:
            return

        queue = self.__queues[key]
        queue.put(record)

    def pop(self) -> Optional[R]:
        if not self.__queues:
            return None

        queue_keys = list(self.__queues.keys())
        num_queues = len(queue_keys)

        for _ in range(num_queues):
            new_key_index = (self.__last_key_index + 1) % num_queues

            key = queue_keys[new_key_index]
            queue = self.__queues[key]

            self.__last_key_index = new_key_index

            try:
                record = queue.get(block=False)
                return record
            except (Empty, ValueError, AttributeError):
                continue

        return None

    def get_total_size(self) -> int:
        return sum([queue.qsize() for queue in self.__queues.values()])

    def is_totally_empty(self) -> bool:
        return all([queue.empty() for queue in self.__queues.values()])



class MessageFqnNotFoundException(Exception):
    def __init__(self, message_fqn: str):
        self.message_fqn = message_fqn
        super().__init__(f"Message with fqn {message_fqn} has not been found")



from dataclasses import dataclass, fields
from datetime import datetime
from typing import Any, ClassVar
from uuid import UUID

DlqRecordId = UUID


@dataclass
class DlqRecord:
    __EXCEPTION_MESSAGE_MAX_CHARACTERS: ClassVar[int] = 300
    DATE_TIME_FORMAT: ClassVar[str] = "%Y-%m-%d %H:%M:%S.%f"

    id: DlqRecordId
    event_id: UUID
    subscriber_fqn: str
    event_payload: dict
    exception_type: str
    exception_message: str
    last_failed_at: datetime

    def __post_init__(self) -> None:
        self.exception_message = self.__add_ellipsis(self.exception_message)

    def __add_ellipsis(self, message: str) -> str:
        if len(message) <= self.__EXCEPTION_MESSAGE_MAX_CHARACTERS:
            return message
        return message[: self.__EXCEPTION_MESSAGE_MAX_CHARACTERS - 3] + "..."

    def set_exception(self, exception: Exception) -> None:
        self.exception_type = type(exception).__name__
        self.exception_message = self.__add_ellipsis(str(exception))

    def mark_as_failed(self) -> None:
        self.last_failed_at = datetime.now()

    def get_attrs(self) -> dict[str, Any]:
        attrs = {}
        for field in fields(self):
            property_name = field.name
            property_value = getattr(self, property_name)
            attrs[property_name] = property_value

        return attrs



from abc import abstractmethod

from buz.event import Event
from buz.event.middleware import PublishCallable, PublishMiddleware


class BasePublishMiddleware(PublishMiddleware):
    def on_publish(self, event: Event, publish: PublishCallable) -> None:
        self._before_on_publish(event)
        publish(event)
        self._after_on_publish(event)

    @abstractmethod
    def _before_on_publish(self, event: Event) -> None:
        pass

    @abstractmethod
    def _after_on_publish(self, event: Event) -> None:
        pass



from __future__ import annotations


STRING_ENCODING = "utf-8"


class KafkaHeaderSerializer:
    def serialize(self, headers: dict[str, str]) -> list[tuple[str, bytes]]:
        return [(key, self.__str_to_bytes(value)) for key, value in headers.items()]

    def deserialize(self, data: list[tuple[str, bytes]]) -> dict[str, str]:
        deserialized_data = [(key, self.__bytes_to_str(value)) for key, value in data]
        return dict(deserialized_data)

    def __str_to_bytes(self, data: str) -> bytes:
        return bytes(data, STRING_ENCODING)

    def __bytes_to_str(self, data: bytes) -> str:
        return data.decode(STRING_ENCODING)



from abc import ABC, abstractmethod
from typing import Sequence

from buz.event.dead_letter_queue.dlq_criteria import DlqCriteria
from buz.event.dead_letter_queue.dlq_record import DlqRecord, DlqRecordId


class DlqRepository(ABC):
    @abstractmethod
    def save(self, dlq_record: DlqRecord) -> None:
        pass

    @abstractmethod
    def find_one_or_fail_by_criteria(self, criteria: DlqCriteria) -> DlqRecord:
        pass

    @abstractmethod
    def delete(self, dlq_record_id: DlqRecordId) -> None:
        pass

    @abstractmethod
    def bulk_delete(self, dlq_record_ids: Sequence[DlqRecordId]) -> None:
        pass



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Union


@dataclass(frozen=True)
class KafkaPollRecord:
    key: Optional[Union[str, bytes]]
    headers: list[tuple[str, bytes]]
    value: Optional[bytes]
    timestamp: int
    partition: int
    topic: str
    offset: int



from dataclasses import dataclass
from buz.event.infrastructure.models.delivery_context import DeliveryContext


@dataclass(frozen=True)
class KafkaDeliveryContext(DeliveryContext):
    topic: str
    consumer_group: str
    partition: int
    timestamp: int



from logging import Logger
from typing import Optional, Sequence, Type, TypeVar

from aiokafka.coordinator.assignors.abstract import AbstractPartitionAssignor

from buz.event import Event, Subscriber
from buz.event.infrastructure.buz_kafka.base_buz_aiokafka_async_consumer import (
    BaseBuzAIOKafkaAsyncConsumer,
)
from buz.event.infrastructure.buz_kafka.consume_strategy.consume_strategy import KafkaConsumeStrategy
from buz.event.infrastructure.buz_kafka.consume_strategy.kafka_on_fail_strategy import KafkaOnFailStrategy
from buz.event.infrastructure.buz_kafka.kafka_event_subscriber_executor import KafkaEventSubscriberExecutor
from buz.event.infrastructure.buz_kafka.kafka_event_sync_subscriber_executor import KafkaEventSyncSubscriberExecutor
from buz.event.meta_subscriber import MetaSubscriber
from buz.event.middleware.consume_middleware import ConsumeMiddleware
from buz.event.strategies.retry.consume_retrier import ConsumeRetrier
from buz.event.strategies.retry.reject_callback import RejectCallback
from buz.kafka.domain.models.auto_create_topic_configuration import AutoCreateTopicConfiguration
from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition
from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.services.kafka_admin_client import KafkaAdminClient
from buz.kafka.infrastructure.deserializers.byte_deserializer import ByteDeserializer
from buz.kafka.infrastructure.deserializers.bytes_to_message_deserializer import BytesToMessageDeserializer
from buz.kafka.infrastructure.deserializers.implementations.json_bytes_to_message_deserializer import (
    JSONBytesToMessageDeserializer,
)
from buz.kafka.infrastructure.serializers.kafka_header_serializer import KafkaHeaderSerializer

T = TypeVar("T", bound=Event)


class BuzAIOKafkaMultiThreadedConsumer(BaseBuzAIOKafkaAsyncConsumer):
    def __init__(
        self,
        *,
        connection_config: KafkaConnectionConfig,
        kafka_admin_client: Optional[KafkaAdminClient],
        consume_strategy: KafkaConsumeStrategy,
        on_fail_strategy: KafkaOnFailStrategy,
        max_queue_size: int,
        max_records_retrieved_per_poll: int,
        kafka_partition_assignors: tuple[Type[AbstractPartitionAssignor], ...] = (),
        subscribers: Sequence[Subscriber],
        logger: Logger,
        consumer_initial_offset_position: ConsumerInitialOffsetPosition,
        deserializers_per_subscriber: dict[Subscriber, BytesToMessageDeserializer[T]],
        worker_instance_id: Optional[str] = None,
        consume_middlewares: Optional[Sequence[ConsumeMiddleware]] = None,
        consume_retrier: Optional[ConsumeRetrier] = None,
        reject_callback: Optional[RejectCallback] = None,
        auto_create_topic_configuration: Optional[AutoCreateTopicConfiguration] = None,
        seconds_between_executions_if_there_are_no_tasks_in_the_queue: Optional[int] = None,
        seconds_between_polls_if_there_are_tasks_in_the_queue: Optional[int] = None,
        seconds_between_polls_if_there_are_no_new_tasks: Optional[int] = None,
        max_number_of_concurrent_polling_tasks: Optional[int] = None,
        session_timeout_ms: Optional[int] = None,
        max_poll_interval_ms: Optional[int] = None,
        heartbeat_interval_ms: Optional[int] = None,
        health_check_port: Optional[int] = None,
        wait_for_connection_to_cluster_ms: Optional[int] = None,
    ):
        super().__init__(
            connection_config=connection_config,
            kafka_admin_client=kafka_admin_client,
            consume_strategy=consume_strategy,
            max_queue_size=max_queue_size,
            max_records_retrieved_per_poll=max_records_retrieved_per_poll,
            kafka_partition_assignors=kafka_partition_assignors,
            subscribers=subscribers,
            logger=logger,
            consumer_initial_offset_position=consumer_initial_offset_position,
            auto_create_topic_configuration=auto_create_topic_configuration,
            seconds_between_executions_if_there_are_no_tasks_in_the_queue=seconds_between_executions_if_there_are_no_tasks_in_the_queue,
            seconds_between_polls_if_there_are_tasks_in_the_queue=seconds_between_polls_if_there_are_tasks_in_the_queue,
            seconds_between_polls_if_there_are_no_new_tasks=seconds_between_polls_if_there_are_no_new_tasks,
            max_number_of_concurrent_polling_tasks=max_number_of_concurrent_polling_tasks,
            health_check_port=health_check_port,
            session_timeout_ms=session_timeout_ms,
            max_poll_interval_ms=max_poll_interval_ms,
            heartbeat_interval_ms=heartbeat_interval_ms,
            wait_for_connection_to_cluster_ms=wait_for_connection_to_cluster_ms,
            worker_instance_id=worker_instance_id,
        )
        self.__on_fail_strategy = on_fail_strategy
        self.__consume_middlewares = consume_middlewares
        self.__consume_retrier = consume_retrier
        self.__reject_callback = reject_callback
        self._deserializers_per_subscriber = deserializers_per_subscriber

    async def _create_kafka_consumer_executor(
        self,
        subscriber: MetaSubscriber,
    ) -> KafkaEventSubscriberExecutor:
        if not isinstance(subscriber, Subscriber):
            raise TypeError(
                f"Subscriber {subscriber.__class__.__name__} is not a subclass of Subscriber, probably you are trying to use an asynchronous subscriber"
            )

        byte_deserializer: ByteDeserializer[Event] = self._deserializers_per_subscriber.get(
            subscriber
        ) or JSONBytesToMessageDeserializer(
            # todo: it looks like in next python versions the inference engine is powerful enough to ensure this type, so we can remove it when we upgrade the python version of the library
            event_class=subscriber.handles()  # type: ignore
        )

        return KafkaEventSyncSubscriberExecutor(
            logger=self._logger,
            byte_deserializer=byte_deserializer,
            header_deserializer=KafkaHeaderSerializer(),
            on_fail_strategy=self.__on_fail_strategy,
            subscriber=subscriber,
            consume_middlewares=self.__consume_middlewares,
            consume_retrier=self.__consume_retrier,
            reject_callback=self.__reject_callback,
        )



from abc import ABC, abstractmethod


class ConnectionManager(ABC):
    @abstractmethod
    def connect(self) -> None:
        pass

    @abstractmethod
    def disconnect(self) -> None:
        pass



class InvalidMaxRetriesParamException(Exception):
    def __init__(self, max_retries: int) -> None:
        self.max_retries = max_retries
        super().__init__(f"Max retries cannot be negative, provided value: {max_retries}")



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class AutoCreateTopicConfiguration:
    partitions: int
    replication_factor: int
    configs: dict[str, Any]



from __future__ import annotations

from dataclasses import dataclass

from buz.kafka.domain.models.kafka_connection_credentials import KafkaConnectionCredentials
from buz.kafka.domain.models.kafka_supported_sasl_mechanisms import KafkaSupportedSaslMechanisms
from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols


@dataclass(frozen=True)
class KafkaConnectionSaslCredentials(KafkaConnectionCredentials):
    security_protocol: KafkaSupportedSecurityProtocols
    user: str
    password: str
    sasl_mechanism: KafkaSupportedSaslMechanisms



from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition


class AIOKafkaConsumerInitialOffsetPositionTranslator:
    @classmethod
    def to_kafka_supported_format(cls, consumer_initial_offset_position: ConsumerInitialOffsetPosition) -> str:
        if consumer_initial_offset_position == ConsumerInitialOffsetPosition.BEGINNING:
            return "earliest"
        if consumer_initial_offset_position == ConsumerInitialOffsetPosition.END:
            return "latest"

        raise ValueError(f"Invalid ConsumerInitialOffsetPosition: {consumer_initial_offset_position}")



from abc import abstractmethod, ABC
from buz.event.meta_subscriber import MetaSubscriber


class KafkaConsumeStrategy(ABC):
    @abstractmethod
    def get_topics(self, subscriber: MetaSubscriber) -> list[str]:
        pass

    @abstractmethod
    def get_subscription_group(self, subscriber: MetaSubscriber) -> str:
        pass



from dataclasses import dataclass
from typing import Any, Optional


@dataclass(frozen=True)
class CreateKafkaTopic:
    name: str
    partitions: int
    replication_factor: int
    configs: Optional[dict[str, Any]] = None



from __future__ import annotations

import json
from datetime import date, datetime, time
from enum import Enum
from typing import Any, Union

from buz.kafka.infrastructure.serializers.byte_serializer import ByteSerializer

STRING_ENCODING = "utf-8"

PRIMITIVES = Union[int, str, float, None, bool, datetime, date, time, Enum]

ALLOWED_SERIALIZABLE_DICTIONARY_KEYS = str

JSON_SERIALIZABLE_VALUE = Union[PRIMITIVES, list[PRIMITIVES], dict[ALLOWED_SERIALIZABLE_DICTIONARY_KEYS, PRIMITIVES]]

JSON_SERIALIZABLE = Union[
    dict[ALLOWED_SERIALIZABLE_DICTIONARY_KEYS, Union[JSON_SERIALIZABLE_VALUE, "JSON_SERIALIZABLE"]]
]


class JSONByteSerializer(ByteSerializer[JSON_SERIALIZABLE]):
    def serialize(self, data: JSON_SERIALIZABLE) -> bytes:
        json_object = self.serialize_as_json(data)
        encoded_string = bytes(
            json_object,
            STRING_ENCODING,
        )

        return encoded_string

    def serialize_as_json(self, data: JSON_SERIALIZABLE) -> str:
        return json.dumps(
            data,
            sort_keys=True,
            default=lambda value: self.__json_serial(value),
        )

    # @see https://stackoverflow.com/questions/11875770/how-can-i-overcome-datetime-datetime-not-json-serializable
    def __json_serial(self, obj: Any) -> Any:
        """JSON serializer for objects not serializable by default json code"""
        if isinstance(obj, (datetime, date, time)):
            return obj.isoformat()
        raise TypeError("Type %s not serializable" % type(obj))



from dataclasses import dataclass
from typing import ClassVar, Union

from buz.event.dead_letter_queue.dlq_record import DlqRecordId


@dataclass(frozen=True)
class DlqCriteria:
    UNSET_VALUE: ClassVar[object] = object()

    dlq_record_id: Union[DlqRecordId, None, object] = UNSET_VALUE



from buz.command.command import Command
from buz.command.more_than_one_command_handler_related_exception import MoreThanOneCommandHandlerRelatedException


__all__ = ["Command", "MoreThanOneCommandHandlerRelatedException"]



from abc import abstractmethod, ABC

from buz.event import Event


class PublishRetryPolicy(ABC):
    @abstractmethod
    def max_retries(self, event: Event) -> int:
        pass

    @abstractmethod
    def interval_start(self, event: Event) -> float:
        pass

    @abstractmethod
    def interval_step(self, event: Event) -> float:
        pass

    @abstractmethod
    def interval_max(self, event: Event) -> float:
        pass

    @abstractmethod
    def error_callback(self, event: Event, exc: Exception, interval_range: range) -> None:
        pass



from typing import Optional, Iterable

from buz.event import Event, EventBus
from buz.event.middleware.publish_middleware import PublishMiddleware
from buz.event.middleware.publish_middleware_chain_resolver import PublishMiddlewareChainResolver
from buz.event.transactional_outbox import OutboxRecord
from buz.event.transactional_outbox.event_to_outbox_record_translator import EventToOutboxRecordTranslator
from buz.event.transactional_outbox.outbox_record_validation.outbox_record_validator import OutboxRecordValidator
from buz.event.transactional_outbox.outbox_repository import OutboxRepository


class TransactionalOutboxEventBus(EventBus):
    def __init__(
        self,
        outbox_repository: OutboxRepository,
        event_to_outbox_record_translator: EventToOutboxRecordTranslator,
        outbox_record_validator: Optional[OutboxRecordValidator] = None,
        publish_middlewares: Optional[list[PublishMiddleware]] = None,
    ) -> None:
        self.__outbox_repository = outbox_repository
        self.__event_to_outbox_record_translator = event_to_outbox_record_translator
        self.__outbox_record_validator = outbox_record_validator
        self.__publish_middleware_chain_resolver = PublishMiddlewareChainResolver(publish_middlewares or [])

    def publish(self, event: Event) -> None:
        self.__publish_middleware_chain_resolver.resolve(event, self.__perform_publish)

    def __perform_publish(self, event: Event) -> None:
        outbox_record = self.__process_event(event)
        self.__outbox_repository.save(outbox_record)

    def bulk_publish(self, events: Iterable[Event]) -> None:
        outbox_records: list[OutboxRecord] = []

        for event in events:
            self.__publish_middleware_chain_resolver.resolve(event, lambda resolved_event: None)
            outbox_records.append(self.__process_event(event))

        if len(outbox_records) == 0:
            return None

        self.__outbox_repository.bulk_create(outbox_records)

    def __process_event(self, event: Event) -> OutboxRecord:
        outbox_record = self.__translate_event(event)
        self.__validate_outbox_record(outbox_record)
        return outbox_record

    def __translate_event(self, event: Event) -> OutboxRecord:
        return self.__event_to_outbox_record_translator.translate(event)

    def __validate_outbox_record(self, outbox_record: OutboxRecord) -> None:
        if self.__outbox_record_validator is not None:
            self.__outbox_record_validator.validate(record=outbox_record)



from __future__ import annotations

from dataclasses import asdict
from datetime import datetime
from typing import Optional

from buz.event import Event
from buz.kafka.infrastructure.cdc.cdc_message import CDCMessage, CDCPayload
from buz.kafka.infrastructure.cdc.cdc_schema import generate_cdc_schema
from buz.kafka.infrastructure.serializers.byte_serializer import ByteSerializer
from buz.kafka.infrastructure.serializers.implementations.json_byte_serializer import JSONByteSerializer


class CDCRecordBytesToEventSerializer(ByteSerializer):
    def __init__(self) -> None:
        self.__json_serializer = JSONByteSerializer()

    def serialize(self, data: Event) -> bytes:
        cdc_message: CDCMessage = CDCMessage(
            payload=CDCPayload(
                event_id=data.id,
                created_at=self.__adapt_created_to_cdc_format(data.created_at),
                event_fqn=data.fqn(),
                payload=self.__serialize_payload(data),
                event_metadata=self.__serialize_event_metadata(data),
            ),
            schema=generate_cdc_schema(data),
        )
        return self.__json_serializer.serialize(asdict(cdc_message))

    def __adapt_created_to_cdc_format(self, created_at: str) -> str:
        created_at_datetime = datetime.strptime(created_at, Event.DATE_TIME_FORMAT)
        return created_at_datetime.strftime(CDCPayload.DATE_TIME_FORMAT)

    def __serialize_payload(self, event: Event) -> str:
        payload = asdict(event)
        del payload["id"]
        del payload["created_at"]
        del payload["metadata"]
        return self.__json_serializer.serialize_as_json(payload)

    def __serialize_event_metadata(self, event: Event) -> Optional[str]:
        return self.__json_serializer.serialize_as_json(event.metadata)  # type: ignore[arg-type]



from buz.query.query import Query
from buz.query.query_response import QueryResponse
from buz.query.more_than_one_query_handler_related_exception import MoreThanOneQueryHandlerRelatedException

__all__ = ["Query", "QueryResponse", "MoreThanOneQueryHandlerRelatedException"]



from dataclasses import dataclass

from buz.event.infrastructure.models.delivery_context import DeliveryContext


@dataclass(frozen=True)
class ExecutionContext:
    delivery_context: DeliveryContext



from __future__ import annotations

import asyncio
from logging import Logger
from typing import Awaitable, Callable, Optional, Sequence, cast

from aiokafka import AIOKafkaConsumer as AIOKafkaNativeConsumer, ConsumerRecord, TopicPartition, OffsetAndMetadata
from aiokafka.helpers import create_ssl_context

from buz.event.infrastructure.buz_kafka.exceptions.kafka_event_bus_config_not_valid_exception import (
    KafkaEventBusConfigNotValidException,
)
from buz.kafka.domain.exceptions.topic_already_created_exception import KafkaTopicsAlreadyCreatedException
from buz.kafka.domain.models.auto_create_topic_configuration import AutoCreateTopicConfiguration
from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition
from buz.kafka.domain.models.create_kafka_topic import CreateKafkaTopic
from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.models.kafka_poll_record import KafkaPollRecord
from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols
from buz.kafka.domain.services.kafka_admin_client import KafkaAdminClient
from buz.kafka.infrastructure.aiokafka.rebalance.kafka_callback_rebalancer import KafkaCallbackRebalancer
from buz.kafka.infrastructure.aiokafka.translators.consumer_initial_offset_position_translator import (
    AIOKafkaConsumerInitialOffsetPositionTranslator,
)


class AIOKafkaConsumer:
    __DEFAULT_POLL_TIMEOUT_MS = 0
    __DEFAULT_ASYNC_FUNCTION_TIMEOUT_MS = 3000

    def __init__(
        self,
        *,
        consumer_group: str,
        topics: list[str],
        connection_config: KafkaConnectionConfig,
        kafka_admin_client: Optional[KafkaAdminClient],
        initial_offset_position: ConsumerInitialOffsetPosition,
        partition_assignors: tuple,
        logger: Logger,
        session_timeout_ms: int,
        max_poll_interval_ms: int,
        heartbeat_interval_ms: int,
        on_partition_revoked: Callable[[AIOKafkaConsumer, set[TopicPartition]], Awaitable[None]],
        on_partition_assigned: Callable[[AIOKafkaConsumer, set[TopicPartition]], Awaitable[None]],
        auto_create_topic_configuration: Optional[AutoCreateTopicConfiguration] = None,
        wait_for_connection_to_cluster_ms: Optional[int] = None,
        group_instance_id: Optional[str] = None,
    ) -> None:
        self.__consumer_group = consumer_group
        self.__topics = topics
        self.__initial_offset_position = initial_offset_position
        self.__connection_config = connection_config
        self.__kafka_admin_client = kafka_admin_client
        self.__partition_assignors = partition_assignors
        self.__logger = logger
        self.__session_timeout_ms = session_timeout_ms
        self.__auto_create_topic_configuration = auto_create_topic_configuration
        self.__on_partitions_revoked_callback = on_partition_revoked
        self.__on_partitions_assigned_callback = on_partition_assigned
        self.__max_poll_interval_ms = max_poll_interval_ms
        self.__heartbeat_interval_ms = heartbeat_interval_ms
        self.__wait_for_connection_to_cluster_ms = wait_for_connection_to_cluster_ms
        self.__group_instance_id = group_instance_id
        self.__check_kafka_admin_client_is_needed()
        self.__consumer = self.__generate_consumer()

    def get_topics(self) -> Sequence[str]:
        return list(self.__topics)

    def get_consumer_group(self) -> str:
        return self.__consumer_group

    def __check_kafka_admin_client_is_needed(self) -> None:
        if self.__kafka_admin_client is None and self.__auto_create_topic_configuration is not None:
            raise KafkaEventBusConfigNotValidException(
                "A KafkaAdminClient is needed to create topics when 'auto_create_topic_configuration' is set."
            )

    def __generate_consumer(self) -> AIOKafkaNativeConsumer:
        if self.__auto_create_topic_configuration is not None:
            self.__ensure_topics_are_created(self.__auto_create_topic_configuration)

        sasl_mechanism = (
            self.__connection_config.credentials.sasl_mechanism.value
            if self.__connection_config.credentials.sasl_mechanism
            else "PLAIN"
        )

        ssl_context = (
            create_ssl_context()
            if self.__connection_config.credentials.security_protocol == KafkaSupportedSecurityProtocols.SASL_SSL
            else None
        )

        consumer = AIOKafkaNativeConsumer(
            None,
            group_instance_id=self.__group_instance_id,
            ssl_context=ssl_context,
            bootstrap_servers=",".join(self.__connection_config.bootstrap_servers),
            security_protocol=self.__connection_config.credentials.security_protocol.value,
            sasl_mechanism=sasl_mechanism,
            sasl_plain_username=self.__connection_config.credentials.user,
            sasl_plain_password=self.__connection_config.credentials.password,
            client_id=self.__connection_config.client_id,
            group_id=self.__consumer_group,
            enable_auto_commit=False,
            auto_offset_reset=AIOKafkaConsumerInitialOffsetPositionTranslator.to_kafka_supported_format(
                self.__initial_offset_position
            ),
            session_timeout_ms=self.__session_timeout_ms,
            heartbeat_interval_ms=self.__heartbeat_interval_ms,
            # partition_assignment_strategy=list(self.__partition_assignors),
            max_poll_interval_ms=self.__max_poll_interval_ms,
            rebalance_timeout_ms=self.__max_poll_interval_ms,
        )

        return consumer

    def __ensure_topics_are_created(self, auto_create_topic_configuration: AutoCreateTopicConfiguration) -> None:
        kafka_admin_client = self.__get_kafka_admin_client()
        non_created_topics = [topic for topic in self.__topics if not kafka_admin_client.is_topic_created(topic)]

        if len(non_created_topics) == 0:
            return None

        topics_to_create = [
            CreateKafkaTopic(
                name=topic,
                partitions=auto_create_topic_configuration.partitions,
                replication_factor=auto_create_topic_configuration.replication_factor,
                configs=auto_create_topic_configuration.configs,
            )
            for topic in non_created_topics
        ]

        try:
            self.__logger.info(f"Creating missing topics: {non_created_topics}...")
            kafka_admin_client.create_topics(topics=topics_to_create)
            self.__logger.info(f"Created missing topics: {non_created_topics}")
        except KafkaTopicsAlreadyCreatedException:
            # there is a possibility to have a race condition between the check and the creation
            # but it does not matters, the important part is that the topic is created
            pass

    def __get_kafka_admin_client(self) -> KafkaAdminClient:
        if self.__kafka_admin_client is None:
            raise KafkaEventBusConfigNotValidException("KafkaAdminClient is not set.")
        return self.__kafka_admin_client

    async def init(self) -> None:
        self.__consumer.subscribe(
            topics=self.__topics,
            listener=KafkaCallbackRebalancer(
                logger=self.__logger,
                on_partitions_assigned=self.__on_partitions_assigned,
                on_partitions_revoked=self.__on_partitions_revoked,
            ),
        )

        self.__logger.info(f"Initializing connection of consumer with group_id={self.__consumer_group}")

        if self.__wait_for_connection_to_cluster_ms is not None:
            await asyncio.wait_for(self.__consumer.start(), self.__wait_for_connection_to_cluster_ms / 1000)
        else:
            await self.__consumer.start()

    async def __on_partitions_assigned(
        self,
        topics_partitions: set[TopicPartition],
    ) -> None:
        await self.__on_partitions_assigned_callback(self, topics_partitions)

    async def __on_partitions_revoked(
        self,
        topics_partitions: set[TopicPartition],
    ) -> None:
        await self.__on_partitions_revoked_callback(self, topics_partitions)

    async def poll(
        self,
        *,
        number_of_messages_to_poll: Optional[int] = None,
    ) -> list[KafkaPollRecord]:
        poll_results = await self.__get_many(
            max_records=number_of_messages_to_poll,
        )

        if poll_results is None:
            return []

        results = [
            cast(KafkaPollRecord, consumer_record)
            for consumer_records in poll_results.values()
            for consumer_record in consumer_records
        ]

        return results

    async def __get_many(
        self,
        max_records: Optional[int] = None,
    ) -> Optional[dict[TopicPartition, list[ConsumerRecord]]]:
        try:
            return await asyncio.wait_for(
                self.__consumer.getmany(timeout_ms=self.__DEFAULT_POLL_TIMEOUT_MS, max_records=max_records),
                timeout=self.__DEFAULT_ASYNC_FUNCTION_TIMEOUT_MS / 1000,
            )
        except asyncio.TimeoutError:
            self.__logger.debug("Timeout while polling")
            return None

    async def commit_poll_record(self, poll_record: KafkaPollRecord) -> None:
        topic_partition = TopicPartition(topic=poll_record.topic, partition=poll_record.partition)
        offset = {topic_partition: OffsetAndMetadata(poll_record.offset + 1, "")}

        await self.__consumer.commit(offset)

    async def stop(self) -> None:
        self.__logger.info(f"Closing connection of consumer with group_id={self.__consumer_group}")

        self.__consumer.unsubscribe()

        try:
            await asyncio.wait_for(
                self.__consumer.stop(),
                timeout=self.__DEFAULT_ASYNC_FUNCTION_TIMEOUT_MS / 1000,
            )
        except asyncio.TimeoutError:
            self.__logger.debug("Timeout while stopping consumer")



from typing import Collection, Union

from buz.query.query import Query
from buz.query.synchronous import QueryHandler
from buz.query.asynchronous import QueryHandler as AsyncQueryHandler


class MoreThanOneQueryHandlerRelatedException(Exception):
    def __init__(self, query: Query, query_handlers: Collection[Union[QueryHandler, AsyncQueryHandler]]):
        self.query = query
        self.query_handlers = query_handlers
        super().__init__(f"There is more than one handler registered for {query.fqn()}.")



from buz.command import Command
from buz.command.asynchronous import CommandHandler
from buz.command.asynchronous.middleware.handle_middleware import HandleMiddleware, HandleCallable
from buz.middleware.middleware_chain_builder import MiddlewareChainBuilder


class HandleMiddlewareChainResolver:
    def __init__(self, middlewares: list[HandleMiddleware]):
        self.__middlewares = middlewares
        self.__middleware_chain_builder: MiddlewareChainBuilder[
            HandleCallable, HandleMiddleware
        ] = MiddlewareChainBuilder(middlewares)

    async def resolve(self, command: Command, command_handler: CommandHandler, handle: HandleCallable) -> None:
        chain_callable: HandleCallable = self.__middleware_chain_builder.get_chain_callable(
            handle, self.__get_middleware_callable
        )
        await chain_callable(command, command_handler)

    @staticmethod
    def __get_middleware_callable(middleware: HandleMiddleware, next_callable: HandleCallable) -> HandleCallable:
        return lambda command, command_handler: middleware.on_handle(command, command_handler, next_callable)



from typing import Any, TypedDict


class Metadata(TypedDict):
    def __init__(self, **kwargs: dict[str, Any]) -> None:  # type: ignore[misc]
        super().__init__()
        for key, value in kwargs.items():
            setattr(self, key, value)



from enum import Enum


class KafkaSupportedSaslMechanisms(Enum):
    PLAIN = "PLAIN"
    GSSAPI = "GSSAPI"
    OAUTHBEARER = "OAUTHBEARER"
    SCRAM_SHA_256 = "SCRAM-SHA-256"
    SCRAM_SHA_512 = "SCRAM-SHA-512"



from abc import ABC, abstractmethod


class AsyncExecutionStrategy(ABC):
    @abstractmethod
    async def start(self) -> None:
        pass

    @abstractmethod
    def request_stop(self) -> None:
        pass



from abc import abstractmethod
from typing import Callable

from buz.event import Event
from buz.middleware import Middleware

PublishCallable = Callable[[Event], None]


class PublishMiddleware(Middleware):
    @abstractmethod
    def on_publish(self, event: Event, publish: PublishCallable) -> None:
        pass



from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class CDCPayload:
    DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

    payload: str  # json encoded
    event_id: str  # uuid
    created_at: str
    event_fqn: str
    event_metadata: Optional[str] = None  # json encoded

    def validate(self) -> None:
        if not isinstance(self.payload, str):
            raise ValueError("The payload value is not a valid value")
        if not isinstance(self.event_id, str):
            raise ValueError("The event_id value is not a valid value")
        if not isinstance(self.created_at, str):
            raise ValueError("The created_at value is not a value")
        if not isinstance(self.event_fqn, str):
            raise ValueError("The event_fqn value is not a valid value")
        if self.event_metadata is not None and not isinstance(self.event_metadata, str):
            raise ValueError("The event_metadata value is not a valid value")

    def __post_init__(self) -> None:
        self.validate()



from __future__ import annotations


class NotAllPartitionAssignedException(Exception):
    def __init__(
        self,
        *,
        topic_name: str,
        consumer_group: str,
    ) -> None:
        super().__init__(
            f'Not all the partitions in the consumer group "{consumer_group}" were assigned in the topic "{topic_name}". Please disconnect the rest of subscribers'
        )



from abc import ABC, abstractmethod
from typing import Sequence

from buz.event import Event
from buz.event.meta_subscriber import MetaSubscriber


class RejectCallback(ABC):
    @abstractmethod
    def on_reject(self, event: Event, subscribers: Sequence[MetaSubscriber], exception: Exception) -> None:
        pass



from buz.command.asynchronous.middleware.base_handle_middleware import BaseHandleMiddleware
from buz.command.asynchronous.middleware.handle_middleware import HandleMiddleware, HandleCallable
from buz.command.asynchronous.middleware.handle_middleware_chain_resolver import HandleMiddlewareChainResolver

__all__ = ["HandleMiddleware", "BaseHandleMiddleware", "HandleMiddlewareChainResolver", "HandleCallable"]



from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar

from buz import Handler
from buz.command import Command

TCommand = TypeVar("TCommand", bound=Command)


class CommandHandler(Generic[TCommand], Handler[TCommand], ABC):
    @classmethod
    @abstractmethod
    def handles(cls) -> Type[TCommand]:
        pass

    @abstractmethod
    def handle(self, command: TCommand) -> None:
        pass
